#ifndef HIT_TRACK_COUNTER_PROCESSOR_H
#define HIT_TRACK_COUNTER_PROCESSOR_H 1

#define N_1D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR 1
#define N_2D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR 0

//LCIO
#include <lcio.h>

#include "ZBinTemplateProcessor.h"


namespace marlintpc
{

    /** Count hits in events and tracks, and tracks per event.
     *  Fill the reults into AIDA histograms.
     *
     *  This Processor is derrived from ZBinTemplateProcessor to have the number of hits per track
     *  binned in z. See also processor parameters of ZBinTemplateProcessor.
     *
     *  <h4>Input - Prerequisites</h4>
     *  \li A collection of reconstructed tracks <br>
     *      To have the hits per track LCIO::TRBIT_HITS has to be set, so the hits are stored.
     *  \li A collection of reconstructed hits
     *
     *  <h4>Output</h4>
     *  \li Histograms with number of hits per track
     *  \li Histograms with number of hits per event
     *  \li Histograms with number of tracks per event
     *
     *  @param InputTracks Name of input Tracks collection (default: TPCTracks)
     *  @param InputHits Name of input TrackerHits collection (default: TPCHits)
     *  @param MaxHitsPerTrack  Maximum value in HitsPerTrack  histogram (default: 20 )
     *  @param MaxHitsPerEvent  Maximum value in HitsPerEvent  histogram (default: 50 )
     *  @param MaxTracksPerEvent  Maximum value in TracksPerEvent  histogram (default: 10 )
     *
     *  @author Martin Killenberg, University of Bonn
     */

    class HitAndTrackCounterProcessor : public ZBinTemplateProcessor<N_1D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR, 
	                                                             N_2D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR>{
			      
    public:
    
	virtual Processor*  newProcessor() { return new HitAndTrackCounterProcessor ; }
    
	HitAndTrackCounterProcessor();

	virtual ~HitAndTrackCounterProcessor();

	virtual void init();

	virtual void processRunHeader( EVENT::LCRunHeader* run ) ;

	virtual void processEvent( LCEvent * evt ) ; 
    
    protected:

     std::string _tracksColName ;    ///< the name of the input tracks collection
     std::string _hitsColName ;    ///< the name of the input hits collection
     
     const std::string _hitsPerTrackHistosName;
     const std::string _hitsPerEventHistosName;
     const std::string _tracksPerEventHistosName;

     /** The index of the template parameter for the hitsPerTrack histos
      *  We want to have these values divided into z bins
      */
     static const unsigned int _hitsPerTrackHistosIndex;
//     /// The index of the template parameter for the hitsPerEvent histos
//     static const unsigned int _hitsPerEventHistosIndex;
//     /// The index of the template parameter for the tracksPerEvent histos
//     static const unsigned int _stracksPerEventHistosIndex;

     /** For the following histograms a z binning does not make sense.
      *  They are handled in the usual way without using the ZBinTemplateProcessor.
      */
     AIDA::IHistogram1D * _hitsPerEventHisto; ///< number of hits per event
     AIDA::IHistogram1D * _tracksPerEventHisto; ///< number of tracks per event

     int _maxHitsPerTrack; ///<  Maximum value in HitsPerTrack  histogram
     int _maxHitsPerEvent; ///<  Maximum value in HitsPerEvent  histogram
     int _maxTracksPerEvent; ///<  Maximum value in TracksPerEvent histogram 
    };
}

#endif // HIT_TRACK_COUNTER_PROCESSOR_H
