#ifndef TPCElectronicsProcessor_h
#define TPCElectronicsProcessor_h 1

//MARLIN
#include <marlin/Processor.h>

//LCIO
#include <lcio.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerRawData.h>

//C++
#include <string>

//GEAR
#include <gear/PadRowLayout2D.h>

//MARLINTPC
#include "TPCVoxel.h"

namespace marlintpc{

/** This processor applies the electronics shaping to the incoming electron signals.
 *  The signal shape is a gaussian which is cut off at +- 3 \f$\sigma\f$. 
 *  The width of the gaussian is defined as \f$\sigma\f$ = rise time / 3.
 *  The electronics amplification is calculated from the number of of electrons and 
 *  the number of ADC counts which correspond to the dynamic range of the ADC.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of TPCVoxel with charges on pads
 *
 *  <h4>Output</h4> 
 *  Collection with TPCRawData
 *
 * @param InputCollectionName Name of input collection, TPCVoxel
 * @param OutputCollectionName Name of output collection, TrackerRawData
 * @param Transient If true the output collection is set transient, i.e. not written to disc
 * @param ReadoutFrequency The frequency of readout electronics in MHz (default 40.).
 * @param RiseTime Signal rise time in ns (default 60.).
 * @param ADCMaxValue Dynamic range of the ADC in ADC counts (default 1023)
 * @param NElectronsDynamicRange Dynamic range of the ADC in electrons (default 1e6)
 * @param VoxelZLength Length of a voxel in ns (default 1.);
 *
 * @author Astrid Muennich, RWTH Aachen
 * @author Thorsten Krautscheid, University of Bonn
 * @author Martin Killenberg, CERN
 */


class TPCElectronicsProcessor : public marlin::Processor {
  
 public:
  
  virtual Processor*  newProcessor() { return new TPCElectronicsProcessor ; }
  
  TPCElectronicsProcessor() ;
  virtual ~TPCElectronicsProcessor();

  /* Instantiate the ADCPulse with the parameters from the steering file.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( lcio::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( lcio::LCEvent * evt ) ; 
  
  /** A nested helper class which holds the raw data of the pulse and represents
   *  the ADC and its shaping amplifier.
   *
   *  The shaping is impelemted as a Gaussian smearing, the left edge 
   *  (-3\f$\sigma\f$ from the centre of the Gaussian) is the time when the 
   *  charge arrives. Afterwards the number of electrons arriving in the time
   *  bins is calculated by integrating the Gaussian within this time bin.
   *
   *  Charge arriving at different times can be added to one pulse. However, there are two
   *  limitations to get propper results:
   *  \li The (start) time of the new charge must not be later than the end of the charge already 
   *      stored in the pulse. In this case the pulses would in reality be spearated. The 
   *      calling code has to check this, otherwise wou will get an exception.
   *  \li The (start) time of the new charge must not be earlyer than the start time of the
   *      existing pulse. Although the pulses might overlap in reality there is the problen that 
   *      the data vector would have to be reorganised. The calling code can simply avoid this 
   *      by sorting the charges by time before adding them. Like this it is also ensured 
   *      that the overlap detection is working (correct end time of all charges that have arrived before).
   *  
   *  For convenience this class also holds the channel number.
   */

  class ADCPulse
  {
    
  public:
    /// The constructor.
    /// @param frequency ADC readout frequency in MHz
    /// @param riseTime The rise time of the pulse in ns
    /// @param nElectronsDynamicRange The number of electrons which correspund to the dynamic range 
    ///                               of the ADC (ADCMaxValue)
    /// @param ADCMaxValue The dynamic range of the ADC in ADC counts
    ADCPulse(double frequency, double riseTime, double nElectronsDynamicRange, int ADCMaxValue);

    double getEndTime() const; ///< Get the end time of the last bin in the pulse (in ns)
    int getStartBin() const {return _startBin;} ///< Returns the bin number of the first data bin in the pulse
    int getChannelNumber() const ; ///< Returns the channel number
    void clear(int newChannelNumber); ///< Clear the data vector, set startBin to -1 and set the new channel number
    void add(double startTime, int nElectrons); ///< Add signal from charge (nElectrons) arriving 
                                                ///< at \c startTime to the pulse
    
    const std::vector<float> & getRawDataVec() const; ///< Provides access to the raw data of the pulse

    /** Returns a TrackerRawData object to be stored in an lcio file.
     *  With the pointer to the object the calling code also gets the ownership.
     *  In constrast to getRawDataVec() starting and trailing zeros are removed.
     *  However, there might be single samples with 0 within the pulse: In case of
     *  two overlapping pulses the intersection region can have less than 1 ADC 
     *  count of charge. This data is not unrealistic, also in real ALTRO data 
     *  these combined pulses exist (if there only because the single 0 in the data
     *  uses less bytes than a new pulse).
     */

    EVENT::TrackerRawData * getLCIORawData() const; ///< Returns a TrackerRawData object to be stored in an lcio file.
                                              ///< With the pointer to the object the 
                                              ///< calling code also get the ownership.

  protected:

    std::vector<float> _rawData; ///< The raw data of the pulse, still in floats to avouid rounding errors
    double _frequency; ///< Readout freuqency of the ADC
    double _riseTime; ///< Rise time  of the pulse
    int _ADCMaxValue; ///< The maximum of the dynamic rance in ADC counts
    double _sqrt2_sigma; ///< \f$\sqrt{2}\sigma\f$ is often needed during calculation, where
    ///< \f$\sigma\f$ = riseTime / 3 is the width of the Gaussian
    int _startBin; ///< The bin number if the first in the data vector
    int _channelNumber;///< The channel in which the pulse is simulated
    double _ADCChargeFactor; ///< The factor to multiply the number of electron per bin to obtain
                             ///< the ADC counts. This is the calibration factor between the number of 
                             ///< primary  electrons and the sum of ADC counts the pulse.
                             ///< It's the same as the factor 
                             ///< between the ADC value of the peak maximum and the number of electrons 
                             ///< in the maximum bin

  };
  

protected:

  /** Struct to sort TPCVoxel of input collection by pad index.
   */
  struct voxelComp {
    bool operator() (const marlintpc::TPCVoxel & left, const marlintpc::TPCVoxel & right) const
    {
       return( left.getPadIndex() == right.getPadIndex()  ?
	       left.getTimeIndex() < right.getTimeIndex() :
	       left.getPadIndex()  < right.getPadIndex() );
    }
  };

  bool _isTransient;

  //Input-/output collection names
  std::string _inputColName;///<Name of input collection
  std::string _outputColName;///<Name of output collection

  double _frequency; ///< Readout freuqency of the ADC
  double _riseTime; ///< Rise time of the signal
  int  _ADCMaxValue; ///< The maximum of the dynamic rance in ADC counts
  int _nElectronsDynamicRange;  ///< The number of electrons which correspond 
                                  ///< to the dynamic range of the ADC
  double _voxelZLength; ///<  Length of a voxel in ns

  // to be implemented later
  //  int _ADCThreshold; // /<Threshold of the ADC
  //  double _cutOff; // /<Cut off value for the integration

  /// The instance of the ADCPulse, as pointer since it can only be initialised in the init()
  ADCPulse *pulse;
  

};

} //end of namespace marlintpc

#endif 
