#include <HitDistributor.h>

//C++
#include <cmath>
#include <complex>

//MARLIN
#include <streamlog/streamlog.h>


namespace marlintpc{

  //Constructor for FixedPadSizeDiskLayout
  HitDistributor::HitDistributor(gear::PadRowLayout2D  const* padLayout, 
                                 double zBinLength) {

    _padLayout     = padLayout;
    _zBinLength    = zBinLength;
    _padLayoutType = _padLayout->getPadLayoutType();
  }


  HitDistributor::~HitDistributor() {}

  // Steering function. The only function accessible from the outside.
  // Takes a hit for input and calls all other functios for the smearing
  // depending on the pad layout of the readout plane.
  std::list<TPCVoxel> HitDistributor::distributeHit(SimTrackerHit* hit, 
						      double sigmaTrans, double sigmaLong){

    //Get hit values in cartesian coordinates
    _xHit = hit->getPosition()[0];
    _yHit = hit->getPosition()[1];   

    //Distribute charge of the hit according to the pad layout type (polar or rectangular geometry)
    switch( _padLayoutType ){
    case gear::PadRowLayout2D::POLAR: {
      std::list<int> padIndexList = getPadsPolar( sigmaTrans );
      if( padIndexList.empty() )
	return std::list<TPCVoxel>();
      else{
        return distributeHitOnPadsPolar( padIndexList, hit, sigmaTrans, sigmaLong );
      }
      break;
    }
    case gear::PadRowLayout2D::CARTESIAN: {
      std::list<int> padIndexList = getPadsRect( sigmaTrans );
      if( padIndexList.empty() )
	return std::list<TPCVoxel>();
      else
        return distributeHitOnPadsRect( padIndexList, hit, sigmaTrans, sigmaLong );
      break;
    }
    default:
      throw gear::Exception("Unknown pad layout type");
    };

    return std::list<TPCVoxel>();    
  }

  // Gets all pads within the smearing radius and stores their indices 
  //in list  for circular pad layout. 
  std::list<int> HitDistributor::getPadsPolar( double sigmaTrans ){

    std::list<int> padIndexList;

    std::complex<double> compHit( _xHit, _yHit );
    double rHit   = std::abs( compHit );
    double phiHit = std::arg( compHit );
    if( phiHit < 0 )
      phiHit = ( 2 * M_PI )+phiHit;

    int hitIndex;

    try{
       hitIndex = _padLayout->getNearestPad( rHit , phiHit );
    }catch(std::out_of_range& e){
      return padIndexList;
    }

    int cursor = hitIndex;
    if( checkDistancePolar( hitIndex, sigmaTrans ) ){

      padIndexList.push_back( hitIndex );
      getPadsOnTheLeftPolar(  padIndexList, hitIndex, sigmaTrans );
      getPadsOnTheRightPolar( padIndexList, hitIndex, sigmaTrans );
    }else{
      return padIndexList;
    }

    //step up
    int row = _padLayout->getRowNumber( cursor ) ;

    while(1) {
      row++;
      if( row < _padLayout->getNRows() ) {
	try{
	  cursor = _padLayout->getNearestPad( _padLayout->getPadCenter( _padLayout->getPadIndex(row,0) )[0] , phiHit);
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistancePolar( cursor, sigmaTrans) ) {
          padIndexList.push_back( cursor );
	  getPadsOnTheLeftPolar(  padIndexList, cursor, sigmaTrans );
	  getPadsOnTheRightPolar( padIndexList, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }
    cursor = hitIndex;
    row = _padLayout->getRowNumber( cursor );

    //step down 
    while(1) {
      row--;
      if( row >= 0 ) {
	try{
	  cursor = _padLayout->getNearestPad(_padLayout->getPadCenter(_padLayout->getPadIndex(row,0))[0], phiHit);
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistancePolar( cursor, sigmaTrans ) ) {
          padIndexList.push_back( cursor );
	  getPadsOnTheLeftPolar(  padIndexList, cursor, sigmaTrans );
	  getPadsOnTheRightPolar( padIndexList, cursor, sigmaTrans );
	}else
	  break;
      }
      else
	break;
    }

    return padIndexList;
  }

  // Gets all pads within the smearing radius and stores their indices 
  // in list for rectangular pad layout. 
  std::list<int> HitDistributor::getPadsRect( double sigmaTrans ){
    std::list<int> padIndexList;

    int hitIndex;
    try{
       hitIndex = _padLayout->getNearestPad( _xHit , _yHit );
    }catch(std::out_of_range& e){
      return padIndexList;
    }

    int cursor = hitIndex;

    if( checkDistanceRect( hitIndex, sigmaTrans ) ){

      padIndexList.push_back( hitIndex );
      getPadsOnTheLeftRect(  padIndexList, hitIndex, sigmaTrans );
      getPadsOnTheRightRect( padIndexList, hitIndex, sigmaTrans );

    }else
      return padIndexList;

    int row = _padLayout->getRowNumber( cursor );

    // step up 
    while(1) {
      row++;
      if( row < _padLayout->getNRows() ) {
	try{
	  cursor = _padLayout->getNearestPad( _xHit , _padLayout->getPadCenter( _padLayout->getPadIndex(row,0) )[1] );
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistanceRect( cursor, sigmaTrans) ) {
          padIndexList.push_back( cursor );
	  getPadsOnTheLeftRect(  padIndexList, cursor, sigmaTrans );
	  getPadsOnTheRightRect( padIndexList, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }

    cursor = hitIndex;
    row = _padLayout->getRowNumber(cursor);

    // step down 
    while(1) {
      row--;
      if( row >= 0 ){
	try{
	  cursor = _padLayout->getNearestPad(_xHit, _padLayout->getPadCenter( _padLayout->getPadIndex(row,0)  )[1] );

	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistanceRect( cursor, sigmaTrans ) ){
	  padIndexList.push_back( cursor );
	  getPadsOnTheLeftRect(  padIndexList, cursor, sigmaTrans );
	  getPadsOnTheRightRect( padIndexList, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }

    return padIndexList;
  }

  // Gets the indices of all pads in the row, which are left from the given pad 
  //and wthin the smearing radius, and adds them to the list which contains
  // all pad indices of hit. (for circular pad layout)
  void HitDistributor::getPadsOnTheLeftPolar(std::list<int> & padList, 
							 int centerIndex, 
							 double sigmaTrans){
    int np = centerIndex;

    while(1){
      np = _padLayout->getLeftNeighbour( np );

      if( checkDistancePolar( np, sigmaTrans) && np != centerIndex )
	padList.push_back( np );     
      else 
        return;
    }
  }

  // Gets the indices of all pads in the row, which are left from the given pad 
  // and wthin the smearing radius, and adds them to the list which contains
  // all pad indices of hit. (for rectangular pad layout)
  void HitDistributor::getPadsOnTheLeftRect( std::list<int> & padList, int centerIndex, double sigmaTrans){
    int np = centerIndex;

    while(1){
      try{
        np = _padLayout->getLeftNeighbour( np );
      }catch(std::exception& e){   //exception handling has to be adjusted when the new gear version is installed!
	return;
      }
      if( checkDistanceRect( np, sigmaTrans ) )
	padList.push_back( np );
      else
	return;
    }
  }

  // Gets the indices of all pads in the row, which are right from the given pad 
  //  and wthin the smearing radius, and adds them to the list which contains
  //  all pad indices of hit. (for circular pad layout)
  void HitDistributor::getPadsOnTheRightPolar( std::list<int> & padList, 
							   int centerIndex, 
							   double sigmaTrans){
    int np = centerIndex;

    while(1){
      np = _padLayout->getRightNeighbour( np );

      if( checkDistancePolar( np, sigmaTrans) && np != centerIndex )
	padList.push_back( np );
      else
        return;
    }
  }

  // Gets the indices of all pads in the row, which are right from the given pad 
  //  and wthin the smearing radius, and adds them to the list which contains
  //  all pad indices of hit. (for rectangular pad layout)
  void HitDistributor::getPadsOnTheRightRect(std::list<int> & padList, int centerIndex, double sigmaTrans){
    int np = centerIndex;

    while(1){
      try{
	np = _padLayout->getRightNeighbour( np );
      }catch(std::exception& e){
	return;
      }
      if( checkDistanceRect( np, sigmaTrans) )
	padList.push_back( np );
      else
	return;
    }
  }

  // Checks whether the distance between the given pad and the hit is lower than the smearing 
  //  range. Returns true when any part of the pad is within the smearing radius, even when it 
  //  is only a little corner of the pad. (for circular pad layout)

  bool HitDistributor::checkDistancePolar(int padIndex, double sigmaTrans )
  {
    static const double SMEAR = 4;

    std::complex<double> hit(_xHit, _yHit);
    double rHit = std::abs(hit);
    double phiHit = std::arg(hit);

    // add 1e-6 mm = 1 nm for rounding errors in case of sigmaTrans = 0
    if ( _padLayout->getDistanceToPad(rHit, phiHit, padIndex) <= (SMEAR*sigmaTrans + 1e-6) )
    {
      return true;
    }

    return false;
  }

  // Checks whether the distance between the given pad and the hit is lower than the smearing 
  //  range. Returns true when any part of the pad is within the smearing radius, even when it 
  //  is only a little corner of the pad. (for rectangular pad layout)
  bool HitDistributor::checkDistanceRect(int padIndex, double sigmaTrans )
  {
    static const double SMEAR = 4;

    // add 1e-6 mm = 1 nm for rounding errors in case of sigmaTrans = 0
    if ( _padLayout->getDistanceToPad(_xHit, _yHit, padIndex) <= (SMEAR*sigmaTrans + 1e-6) )
    {
      return true;
    }

    return false;
  }

  //Smears the distributes the charge of the given hit in voxels according to the given pads, 
  //transversal and longitudinal diffusion for circular pad layout. 
  std::list<TPCVoxel> HitDistributor::distributeHitOnPadsPolar( std::list<int> padIndexList, SimTrackerHit* hit, 
								  double sigmaTrans, double sigmaLong){

    std::list<TPCVoxel> voxelList;

    if( hit->getdEdx() > 0 ) {

      while(padIndexList.begin()!=padIndexList.end())  {

        int i = padIndexList.front();

        //claculate the x,y coordinates of the pad center
        double xPad = _padLayout->getPadCenter(i)[0]*cos(_padLayout->getPadCenter(i)[1]);
        double yPad = _padLayout->getPadCenter(i)[0]*sin(_padLayout->getPadCenter(i)[1]);

	streamlog_out( DEBUG ) << "get pad coordinates " << std::endl;

        //rotate coordinate system so that the new x,y axis are parallel to pad widht,height
        double xHitRot =   _xHit*sin(_padLayout->getPadCenter(i)[1]) 
                         - _yHit*cos(_padLayout->getPadCenter(i)[1]);

        double yHitRot =    _xHit*cos(_padLayout->getPadCenter(i)[1]) 
                          - _yHit*sin(_padLayout->getPadCenter(i)[1]);
      

        double xPadRot =    xPad*sin(_padLayout->getPadCenter(i)[1]) 
                          - yPad*cos(_padLayout->getPadCenter(i)[1]);

        double yPadRot = +  xPad*cos(_padLayout->getPadCenter(i)[1]) 
                          - yPad*sin(_padLayout->getPadCenter(i)[1]);

	streamlog_out( DEBUG ) << "performed coordinate transformation " << std::endl;
      
        double halfPadPitch  = _padLayout->getPadCenter(i)[0]*_padLayout->getPadPitch(i);/*_padLayout->getFixedPadPitch()/2;*/
        double halfRowHeight = _padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;

	streamlog_out(DEBUG) << "calculated pad heigth and width: " << halfRowHeight << "  " << halfPadPitch << std::endl;

       //start and end values for the integration
	double xEnd  =xPadRot+halfPadPitch;
	double xStart=xPadRot-halfPadPitch;
	double yEnd  =yPadRot+halfRowHeight;
	double yStart=yPadRot-halfRowHeight;

        // integration over a 2d gaussian
        double charge= hit->getdEdx()/4 *((erf((xEnd  - xHitRot)/(sqrt(2.)*sigmaTrans))
					   -erf((xStart- xHitRot)/(sqrt(2.)*sigmaTrans)))
					  * (erf((yEnd  - yHitRot)/(sqrt(2.)*sigmaTrans))	
					     -erf((yStart- yHitRot)/(sqrt(2.)*sigmaTrans))));
	//	if (charge < 0.5) streamlog_out(MESSAGE) << "Charge is " <<charge << ". Integration will result in 0" << std::endl;

	streamlog_out(DEBUG) << "performed integration" << std::endl;

	if ( charge > 0.5 )
	{
	  int zBin = static_cast<int>( hit->getTime()  /_zBinLength ); 

	  if( sigmaLong ){

	    const double tDiff = 4;

	    int minBin = static_cast<int>( (hit->getTime() - (tDiff*sigmaLong)) / _zBinLength  );
	    int maxBin = static_cast<int>( ceil( (hit->getTime() + (tDiff*sigmaLong)) / _zBinLength  ) );

	    if( minBin < 0 )
	      minBin = 0;

	    for(int j = minBin; j <= maxBin; j++){ 
	      
	    int tDiffCharge = static_cast<int> ( round( charge / 2  * ( erf(((j+1)*_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))-
									erf(( j   *_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))  )));
	    
	    voxelList.push_back( TPCVoxel(i,j,tDiffCharge) );
	    }
	  }// if sigmaLong
	  else{ 
	    voxelList.push_back( TPCVoxel(i,zBin,static_cast<int>(round(charge)))); 
	  }
	}// if charge > 0.5

	// always erase the iterator, even if no charge has been deposited
	padIndexList.erase( padIndexList.begin() );
      }
    }
    return voxelList;
  }

  // Smears the distributes the charge of the given hit in voxels according to the given pads, 
  //transversal and longitudinal diffusion for rectangular pad layout. 
  std::list<TPCVoxel> HitDistributor::distributeHitOnPadsRect( std::list<int> padIndexList, SimTrackerHit* hit, 
								 double sigmaTrans, double sigmaLong){

    std::list<TPCVoxel> voxelList;

    if( hit->getdEdx() > 0 ){

      //integration for all the stored pad indices
      while(padIndexList.begin()!=padIndexList.end()) {

        int i = padIndexList.front();

	//calculate the width of the pad (with pitch!)
	int row = _padLayout->getRowNumber( i );
	double halfPadPitch = _padLayout->getPadPitch(i)/2;
	if( _padLayout->getPadsInRow( row ).size() >= 2) {
	  int one = _padLayout->getPadIndex( row, 0);
	  int two = _padLayout->getPadIndex( row, 1);
 	  halfPadPitch = fabs( _padLayout->getPadCenter(one)[0] - _padLayout->getPadCenter(two)[0] ) /2.;
	}

        //start and end values for the integration
        double xStart, xEnd, yStart, yEnd; 
        if( _padLayout->getPadCenter(i)[0] < _xHit ) {
          xStart = _padLayout->getPadCenter(i)[0]+ halfPadPitch;
    	  xEnd   = _padLayout->getPadCenter(i)[0]- halfPadPitch;
        }else {
          xStart = _padLayout->getPadCenter(i)[0]- halfPadPitch;
	  xEnd   = _padLayout->getPadCenter(i)[0]+ halfPadPitch;
        }  
        if( _padLayout->getPadCenter(i)[1]<hit->getPosition()[1]) {
          yStart = _padLayout->getPadCenter(i)[1]+_padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;
	  yEnd   = _padLayout->getPadCenter(i)[1]-_padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;
        }else {
	  yStart = _padLayout->getPadCenter(i)[1]-_padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;
	  yEnd   = _padLayout->getPadCenter(i)[1]+_padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;
        }

        // integration over a 2d gaussian
        double charge= hit->getdEdx()/4 *((erf((xEnd  - _xHit)/(sqrt(2.)*sigmaTrans))
					   -erf((xStart- _xHit)/(sqrt(2.)*sigmaTrans)))
					  * (erf((yEnd  - _yHit)/(sqrt(2.)*sigmaTrans))	
					     -erf((yStart- _yHit)/(sqrt(2.)*sigmaTrans))));


	if ( charge > 0.5 ) // only perform integration if there can be charge after integration and rounding
	{
	  int zBin = static_cast<int>( hit->getTime()  /_zBinLength ); 
	  
	  if( sigmaLong )
	  {
	    const double tDiff = 4;

	    int minBin = static_cast<int>( (hit->getTime() - (tDiff*sigmaLong)) / _zBinLength  );
	    int maxBin = static_cast<int>( ceil( (hit->getTime() + (tDiff*sigmaLong)) / _zBinLength  ));

	    if( minBin < 0 )
	      minBin = 0;

	    for(int j = minBin; j <= maxBin; j++)
	    { 
	      
	      int tDiffCharge = static_cast<int> ( round (charge * ( erf(((j+1)*_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))-
								     erf(( j   *_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))  )));


	      voxelList.push_back( TPCVoxel(i,j,tDiffCharge) );
	    }
	  }
	  else // if sigmalong
	  { 
	    voxelList.push_back( TPCVoxel(i,zBin,static_cast<int>(round(charge)))); 
	  }
	} // if ( charge > 0.5 ) 

	padIndexList.erase(padIndexList.begin());
      }// while  padIndexList
    }
    return voxelList;

  }



} //namespace marlintpc
