/** Processor to sort primary ions into voxel
 */
#include "IonsInVoxelsProcessor.h"

//C++
#include <iostream>
#include <cmath>
#include <complex>

//LCIO
#include <EVENT/SimTrackerHit.h>

//GEAR
#include <gear/TPCParameters.h>

//MARLIN
#include <marlin/Global.h>


using namespace lcio ;
using namespace marlin ;
using namespace std;

namespace marlintpc{

  
IonsInVoxelsProcessor aIonsInVoxelsProcessor ;


IonsInVoxelsProcessor::IonsInVoxelsProcessor() : Processor("IonsInVoxelsProcessor")  {
  
  _description = "IonsInVoxelsProcessor puts primary ions in voxels..." ;
  
  
  // register steering parameters: name, description, class-variable, default value


  registerInputCollection(    LCIO::SIMTRACKERHIT,
                              "InputCollectionName" , 
			      "e- from PrimaryIonisation (default: PrimaryIonisation)"  ,
			      _inputIonsColName ,
			      string("PrimaryIonisation") ) ;

  registerOutputCollection(   LCIO::LCGENERICOBJECT ,
                              "OutputCollectionName" ,
                              "Name of the Output GEM collection (default: IonOccupancy)" ,
                              _outputVoxelColName ,
			      string("IonOccupancy") );

  registerProcessorParameter( "BinLength" , 
			      "Length of a z bin (default 0)"  ,
			      _binLength,
			      (double) 2.5) ;

  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection (default: 0)"  ,
			      _transient,
			      (int) 0) ;

  registerProcessorParameter( "OutputType" , 
			      "Determines if output describes only the ions of the event or all ions in the chamber (default: 0)"  ,
			      _outputType,
			      (int) 0) ;


  _voxelTPC = 0;  
  _padLayout = 0; 
  _padLayoutType = 0; 

}

void IonsInVoxelsProcessor::init() { 

  streamlog_out(MESSAGE2) << _description << endl;

  printParameters() ;

  //Get the pad layout from gear for the later binning of the ions
  //in the x,y-plane.
  try{
    _padLayout = &(Global::GEAR->getTPCParameters().getPadLayout());
    _padLayoutType = _padLayout->getPadLayoutType();
  }
  catch(std::exception& e){
    streamlog_out(ERROR)<< e.what() << std::endl;
    return;
  }

  _maxDriftLength = Global::GEAR->getTPCParameters().getMaxDriftLength();

  //Instanciate a VoxelTPC to store the ions in voxels.
  if( _voxelTPC )
    delete _voxelTPC;
  _voxelTPC = new marlintpc::VoxelTPC( _padLayout, _binLength );

}



void IonsInVoxelsProcessor::processRunHeader( LCRunHeader* run) {

  run->parameters().setValue( _processorName + "_revision", "$Rev: 1410 $" );

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    if ( ! i->second->isOptional() || i->second->valueSet() ) {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }

} 

 
void IonsInVoxelsProcessor::processEvent( LCEvent * evt ) {


  //Gets the input collection of primary ions..
  LCCollection* inputIonCol = NULL;
  try{
    inputIonCol = evt->getCollection( _inputIonsColName );
  }
  catch(DataNotAvailableException& e){
    streamlog_out(WARNING4) << "Could not get input collection " 
			    << _inputIonsColName << endl;
    return;
  }

  //Loops over all ions in the input collection. If the x,y coordinates
  //of the ions are within the pad plane there charge is put in a voxel.
  for(int i = 0; i < inputIonCol->getNumberOfElements() ; i++){

    SimTrackerHit* myIon = dynamic_cast<SimTrackerHit*>( inputIonCol->getElementAt(i) );

    //Set padIndex to the index of the pad nearest to the hit of myIon. 
    //If the hit is outside of the pad plane it is set to minus one.
    int padIndex = getPad( _padLayout , myIon );

    if( padIndex >= 0 && myIon->getPosition()[2] >= 0) {

      int charge = 0;

      if( myIon->getdEdx() > 1 )
	charge = static_cast<int>( myIon->getdEdx() );
      else
	charge = 1;

      double zBinning = _maxDriftLength - myIon->getPosition()[2];

      _voxelTPC->putChargeInVoxel( padIndex , zBinning , charge );  
    }
            
  }
  LCCollectionVec* outputVoxelCol = 0;

  if( _outputType )
    outputVoxelCol = _voxelTPC->getVoxelCollection();
  else
    outputVoxelCol = _voxelTPC->fetchVoxelCollection();

  if ( !outputVoxelCol ){
    streamlog_out(DEBUG2) << "Empty voxel collection while input was not empty." << std::endl;
    return;
  }

  //Set write mode.
  if( _transient ) 
    outputVoxelCol->setTransient(true);

  //Without giving the binnig in z the voxels cannot be used properly.
  float binLength=static_cast<float>(_binLength);
  outputVoxelCol->parameters().setValue("zBinLength",binLength);

  //Adds a collection with all voxelized ions in the TPC to the event.
  evt->addCollection( outputVoxelCol , _outputVoxelColName );

  //if( evt->getEventNumber() % 100 == 0 )
    streamlog_out(MESSAGE3) << "        Event " << evt->getEventNumber() << " done!" << endl;
}

int IonsInVoxelsProcessor::getPad(gear::PadRowLayout2D const* padLayout , SimTrackerHit* hit){

  //First and second coordinate of the hit. They will be set to x,y or 
  //or r,phi depending on the pad layout.
  double hitFirst;
  double hitSecond;

  //The type of ppad layout is used to determine whether
  //the coordinates of the hit have to be in x,z or in 
  //r,phi.
  int padLayoutType = padLayout->getPadLayoutType();

  //Sets hitFirst and hitSecond depending on the pad layout.
  switch( padLayoutType ) {
  case gear::PadRowLayout2D::CARTESIAN:{
    hitFirst  = hit->getPosition()[0];
    hitSecond = hit->getPosition()[1];
    break;
  }
  case gear::PadRowLayout2D::POLAR: {
    complex<double> compHit( hit->getPosition()[0], hit->getPosition()[1] );
    hitFirst = std::abs(compHit);
    hitSecond = std::arg(compHit);
    if( hitSecond < 0 )
      hitSecond = (2*M_PI)+hitSecond;
    break;
  }
  default:
    throw gear::Exception("Unknown pad layout");
  };

  //Return value of the function. Negative values indicate a hit not 
  //on the padplane.
  
  int padIndex = -1;

  //Checks whether hit is on pad plane. PadRowLayout2D.getPlaneExtend() 
  //return x,y or r,phi coordinates depending on the pad layout.
   
  if( hitFirst  > padLayout->getPlaneExtent()[0] &&
      hitFirst  < padLayout->getPlaneExtent()[1] &&
      hitSecond > padLayout->getPlaneExtent()[2] &&
      hitSecond < padLayout->getPlaneExtent()[3]    ) {
    try{
      //Get the pad nearest to the hit
      padIndex = padLayout->getNearestPad( hitFirst , hitSecond );
    }
    catch( std::out_of_range& e ){
      //If try does not work set padIndex to minus one to prevent
      //crashing of the processor.    
      padIndex = -1;
    }
  }
  else
    padIndex = -1;

  //Returns pad index. This should have been set to a negative value if  
  //the hit is not on pad plane or try did not work.
  return padIndex;
}

} // end of namespace marlintpc
