//*****************************************************
// Part of program to simulate a TPC with GEM readout *
//                                                    *
// Author: Astrid Muennich                            *
//                                                    *
//   Astrid Muennich                                  *
//   RWTH Aachen, III. Phys. Inst. B                  * 
//   Office: 28 A 112, Phone: +49/(0)241/8            *
//   Email: muennich@physik.rwth-aachen.de            * 
//                                                    *
//*****************************************************

/** Programm to apply electronic effects like shaping and binning of chareg into ADCCounts.
 */

#include "TPCElectronicsProcessor.h"

//MarlinTPC
#include "TPCVoxel.h"
#include "VoxelTPC.h"

//tpcconddata
#include "Pedestal.h"

//LCIO
//#include <IO/LCWriter.h>
#include <EVENT/LCCollection.h>
#include <IMPL/TrackerRawDataImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCRunHeaderImpl.h>

//Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd/DataFileHandler.hh>

//C++
#include <iostream>
#include <cmath>
#include <set>
#include <cstdlib>



using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace tpcconddata;

namespace marlintpc{


TPCElectronicsProcessor aTPCElectronicsProcessor;

TPCElectronicsProcessor::TPCElectronicsProcessor() : Processor("TPCElectronicsProcessor") 
{
  
  _description = "TPCElectronicsProcessor takes care of binning the times and charges of electrons into an ADC.";
  
  registerInputCollection(    LCIO::LCGENERICOBJECT,
                              "InputCollectionName" , 
			      "Name of the Input collection with TPCVoxels (default: DistributedElectrons)"  ,
			      _inputColName,
			      std::string("DistributedElectrons") ) ;

  registerOutputCollection(   LCIO::TRACKERRAWDATA,
                              "OutputCollectionName",
                              "Name of the output collection with adc values (default: TPCRawData)",
                              _outputColName,
                              std::string("TPCRawData") ),
  
  registerProcessorParameter( "ReadoutFrequency" , 
			      "Readout Frequency in Mhz (default: 40)"  ,
			      _frequency ,
			      (double)40) ;
  
  registerProcessorParameter( "RiseTime" , 
			      "Rise Time of Shaper in ns (default: 60)"  ,
			      _riseTime ,
			      (double)120.) ;
  
  registerProcessorParameter( "ADCMaxValue" , 
			      " Dynamic range of the ADC in ADC counts (default 1023)"  ,
			      _ADCMaxValue ,
			      (int)1023) ;
  
   registerProcessorParameter( "NElectronsDynamicRange",
			       "Dynamic range of the ADC in electrons (default 1e6)",
			       _nElectronsDynamicRange,
			       (int)1e6);
  
//  registerProcessorParameter( "ADCThreshold",
//			      "Threshold for ADC in ADC bins (default: 5)",
//			      _ADCThreshold,
//			      (int)5);

//   registerProcessorParameter( "IntCutOff",
//			       "Integration Cut Off of Gaussian Shape (default: 3.0)",
//			       _cutOff,
//			       (double)3.0);

   registerProcessorParameter( "VoxelZLength",
			       " Length of a voxel in ns (default 1.)",
			       _voxelZLength,
			       (double)1.);

  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection (default: 0)"  ,
			      _isTransient,
			      (bool)false) ;
}

TPCElectronicsProcessor::~TPCElectronicsProcessor()
{
  // delete the pulse which was allocated in init()
  delete pulse;
}

void TPCElectronicsProcessor::init()
{
  streamlog_out(MESSAGE2) << _description << endl;
  printParameters() ;

  pulse = new ADCPulse(_frequency, _riseTime, _nElectronsDynamicRange, _ADCMaxValue);
}


void TPCElectronicsProcessor::processRunHeader( LCRunHeader* run) 
{ 
  run->parameters().setValue( _processorName + "_revision", "$Rev: 1876 $" );

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    if ( ! i->second->isOptional() || i->second->valueSet() ) {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }

} 


void TPCElectronicsProcessor::processEvent( LCEvent * evt ) {

  streamlog_out(DEBUG4) << "Hi, it's me again "<< evt->getEventNumber() << std::endl;

  LCCollectionVec* inputColVec = 0;

  try {
    inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
  }catch (DataNotAvailableException &e) {

    // the collection of electron clouds with that name wasn't in the LCIO event file, skip simulating this event
    //    throw marlin::SkipEventException( this );
   
    return; // just return, otherwise the data from the previous processors is not written since the 
            //  OutputProcessor is also skipped
  }

  
  //Initialise and set transient for output collection
  LCCollectionVec* outputColVec = new LCCollectionVec( LCIO::TRACKERRAWDATA )  ;
  if( _isTransient )
    outputColVec->setTransient(true);

  //Sort TPCVoxels by padIndex
  std::set<TPCVoxel, voxelComp> sortedVoxels;
  
  for( LCCollectionVec::iterator voxelIter = inputColVec->begin() ; 
       voxelIter < inputColVec->end() ;
       voxelIter++ ){

    sortedVoxels.insert( TPCVoxel(*voxelIter) );
  }

  // initialise the pulse with the first pad index
  pulse->clear( sortedVoxels.begin()->getPadIndex() );

  for( std::set<TPCVoxel, voxelComp>::iterator voxelIter = sortedVoxels.begin();
       voxelIter != sortedVoxels.end() ; voxelIter++)
  {
    if ( ( voxelIter->getPadIndex() != pulse->getChannelNumber() )
	 // The charge in this voxel belongs to a new pulse, since it is on another channel

	 || ( (pulse->getStartBin()>=0) && 
	      (voxelIter->getTimeIndex()*_voxelZLength > pulse->getEndTime() ) ) )
         // The pulse is not empty and the voxel time is after the end time of the pulse,
         // so they don't overlap
    {
      // Write the old pulse, then initialise a new one
      TrackerRawData * lcioPulse = pulse->getLCIORawData();
      if ( lcioPulse ) // only add non-empty pulses
      {
	outputColVec->addElement( lcioPulse );
      }
      else
      {
	streamlog_out(DEBUG2) << "empty pulse arrived in processor" << std::endl;
      }

      pulse->clear( voxelIter->getPadIndex() );
    }
    
    pulse->add( voxelIter->getTimeIndex()*_voxelZLength ,  voxelIter->getCharge() );

  }// for voxelIter

  //Add ouput collection to event
  if ( outputColVec->empty() ) 
    delete outputColVec;
  else
    evt->addCollection( outputColVec , _outputColName) ;
       
}

  

}//end of namespace marlintpc
