#ifndef MAPPING_HANDLER_H
#define MAPPING_HANDLER_H

#include <iostream>
#include <ADCChannelMapping.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCGenericObjectImpl.h>


namespace marlintpc{

  /** A helper class for convenient usage of the channel mapping information.
   *  Allows to get the channelID (hardware channel number/ cellID)
   *  if the padID (logical cellID / padIndex from GEAR) is known;
   *  and vice versa.
   *
   *  @author Martin Killenberg, University of Bonn
   */
class MappingHandler
{
 public:
  MappingHandler();

  /** Set a new channel map.
   *  Internally the vectors to access channelID from padID
   *  and vice versa are recalculated.
   */
  void setChannelMap(EVENT::LCCollection *chanMapCol);

  /** Get the channelID and readoutGroup if the padID and ModuleID are known.
   *  The first member in the pair is the channelID, the second the readoutGroup.
   */
  std::pair<int, int> getHardwareID(int padID, int moduleID) const;

  //  /// Get the readoutGroup if the padID and ModuleID are known.
  // int getReadoutGroup(int padID, int moduleID);

  /** Get the padID and moduleID if the channelID and readoutGroup are known.
   *  The first member in the pair is the padID, the second the moduleID.
   */
  std::pair<int,int> getSoftwareID(int channelID, int readoutGroup) const;

  //  /// Get the moduleID if the channelID and readoutGroup are known.
  //  int getModuleID(int channelID, int readoutGroup);


 private:

  /** A nested helper class for sorting 
   *  ADCChannelMapping objects according to software IDs
   */
  class  sortSoftwareID
  {
    public:
      bool operator()(EVENT::LCObject* adccm1, EVENT::LCObject* adccm2)
      {
	tpcconddata::ADCChannelMapping cm1(adccm1);
	tpcconddata::ADCChannelMapping cm2(adccm2);
	
	// if both module IDs are equal sort by pad index
	if ( cm1.getModuleID() == cm2.getModuleID() )
	{
	  return ( cm1.getPadID() < cm2 .getPadID() ) ;
	}
	else // sort by module ID
	{
	  return ( cm1.getModuleID() < cm2.getModuleID() );
	}
      }
  };

  /** A nested helper class for sorting 
   *  ADCChannelMapping objects according to hardware IDs
   */
  class  sortHardwareID
  {
    public:
      bool operator()(EVENT::LCObject* adccm1, EVENT::LCObject* adccm2)
      {
	tpcconddata::ADCChannelMapping cm1(adccm1);
	tpcconddata::ADCChannelMapping cm2(adccm2);
	
	// if both readout groups are equal sort by channel ID
	if ( cm1.getReadoutGroup() == cm2.getReadoutGroup() )
	{
	    return ( cm1.getChannelID() < cm2 .getChannelID() ) ;
	}
	else // sort by module ID
	{
	  return ( cm1.getReadoutGroup() < cm2.getReadoutGroup() );
	}
      }
  };

  /** A nested helper class for comparing two ints to an
   *  ADCChannelMapping object's channelID and readoutGroup
   */
  class  cmpHardwareID
  {
    public:
    bool operator()(EVENT::LCObject* adccm, std::pair<int, int> hardwareID)
    {
      tpcconddata::ADCChannelMapping cm(adccm);
      // if readout groups are equal
      if (cm.getReadoutGroup() == hardwareID.second )
      { // compare channel ids
	return  ( cm.getChannelID() < hardwareID.first );
      }
      else 
      { // compare readout groups
	return ( cm.getReadoutGroup() < hardwareID.second );
      }
    }
  };

  /** A nested helper class for comparing two ints to an
   *  ADCChannelMapping object's padID and moduleID
   */
  class  cmpSoftwareID
  {
    public:
      bool operator()(EVENT::LCObject* adccm, std::pair<int, int> softwareID)
      {
	tpcconddata::ADCChannelMapping cm(adccm);
	// if module IDs are equal
	if (cm.getModuleID() == softwareID.second )
	{ // compare ad ids
	  return  ( cm.getPadID() < softwareID.first );
	}
	else 
	{ // compare module IDs
	  return ( cm.getModuleID() < softwareID.second );
	}
      }
  };

 private:
  IMPL::LCCollectionVec chanMapSortedHardwareID; ///< The mapping information sorted by channelID and readout group
  IMPL::LCCollectionVec chanMapSortedSoftwareID; ///< The mapping information sorted by pad and moduleID

};


}// namespace marlintpc

#endif // MAPPING_HANDLER_H
