#ifdef USE_LCCD

#ifndef PHOTODOTRECONSTRUCTIONPROCESSOR_H_
#define PHOTODOTRECONSTRUCTIONPROCESSOR_H_ 1

//Marlin
#include <marlin/Processor.h> // Integral for any processor

//gear
#include <gear/PadRowLayout2D.h>

//lcio
#include <IMPL/TrackerPulseImpl.h>

namespace marlintpc {


	/** Analyses the effect of variable magnetic field on photodot positions
	 *  
	 *  PhotodotReconstructionProcessor uses the electron cloud from a laser pulse
	 *  in the TPC to examine the effect of the magnetic field.
	 *  Locations of the centres of images of photodots on the readout
	 *  electronics are compared with the positions of dots on the cathode.
	 *
	 * <h4>Prerequisites</h4>
	 * A TrackerPulse collection containing charges from a number of photodots
	 *
	 * @author P. Conley (UVic)<br>
	 */
 	class PhotodotReconstructionProcessor : public marlin::Processor {

	public: // Functions for interaction with Marlin
		
		/** Functions used directly by Marlin
		 */
		virtual Processor * newProcessor() { return new PhotodotReconstructionProcessor ; }
		PhotodotReconstructionProcessor();
		~PhotodotReconstructionProcessor();

		/** Called at the beginning of the job to initialise the processor
		 */
		virtual void init();

		/** Called for every run
		 */
		virtual void processRunHeader( lcio::LCRunHeader * run );

		/** Called for every event - the body of the processor
		 */
		virtual void processEvent( lcio::LCEvent * evt );

		/** Called after the event has been processed
		 */
		virtual void check( lcio::LCEvent * evt );

		/** Called after data processing for clean-up
		 */
		virtual void end();

	private:

		/** Structure holding the information about photodots from the
		 *  geometry_file. The diameter won't be used, but exists in the data.
		 */
		struct Photodot {
			// Cartesian constructor
			Photodot( double x, double y, double diameter ) :
				_x(x),
				_y(y),
				_r( sqrt( x*x + y*y ) ),
				_phi( atan2( y, x ) ),
				_diameter( diameter ) {};
			// Polar constructor
			/* The extra integer argument serves simply to differentiate the two
			 * constructor types. Its value is irrelevant.
			 */
			Photodot( double r, double phi, double diameter, int alt_type ) :
				_x( r * cos( phi ) ),
				_y( r * sin( phi ) ),
				_r(r),
				_phi(phi),
				_diameter( diameter ) {};
			double _x;
			double _y;
			double _r;
			double _phi;
			double _diameter;
		};

		/** Functions to keep the code clean while finding the hit centre
		 */

		/** Find the hit nearest the position of a photodot, by comparing distances
		 *  to pulses.
		 *  Returns a cellID inside the hit.
		 *  Operates on a photodot
		 */
		int findNearestHit( Photodot thisPhotodot );

		/** Find the centre of a hit containing a given pulse, under the assumption
		 *  that deposited charge displays a Gaussian profile, and pads at the centre
		 *  of a hit receive the greatest charge.
		 *  Returns the cellID of the photodot at the bottom of the centre column.
		 *  Operates on a pulse somewhere in a hit, and a container of pulses.
		 */
		int findCentreColumn( int edgePulse );

		/** Find the precise co-ordinate of the centre of a hit.
		 *  Returns the coordinates of the hit centre.
		 *  Operates on a pulse at the bottom of the centre column of a hit and a
		 *  collection of pulses.
		 */
		gear::Vector2D findHitCentre( int columnCentre );

		/** Determine which pulses arrived at the readout electronics as part of
		 *  the clouds resulting from the laser pulse.
		 *  No returned value (operations performed on input collection)
		 *  Operates on the the photodots and input pulses
		 */
		void findPhotodotPulses();

		/** Calculate the total charge from multiple pulses incident on one pad.
		 *  This isn't an optimal solution, as there should be only one valid
		 *  pulse on a pad, but there's no obvious way around the problem.
		 *  Returns the sum of all charges on the pad.
		 *  Operates on the index of a pad.
		 */
		double getPadCharge( int padIndex );

		/** The name of the file containing the positions of the photodots.
		 */
		std::string _geometry_filename;

		/** The name of the collection containing the photoelectric tracker pulses.
		 */
		std::string _input_pulses_collection_name;

		/** The name of the file containing photoelectron position differences.
		 */
		std::string _compared_photodots_filename;

		/** The positions of the photodots.
		 *  This vector is built from the _geometry_filename, so is not initialised
		 *  with the processor
		 */
		std::vector< Photodot > _photodots;
		std::vector< Photodot > _expected_photodots;

		/** Map describing the pulses in the event, sorted by cell ID.
		 */
		std::multimap< int, EVENT::TrackerPulse * > * _pulsesByIndex;

		/** The geometry type of the file in _geometry_filename
		 */
		std::string _geometry_type;

		/** Data describing the layout of an endcap
		 */
		int _padLayoutType;
		gear::PadRowLayout2D const * _padLayout;

		/** Diffusion of the photodots through the TPC.
		 */
		double _diffusion;

		/* The following parameters describe the geometry of the TPC
		 */
		/** The gas parameterisation of this piece of the TPC
		 */
		double _omega_tau;

		/** The paramaterised drift velocity in this section of the TPC (mm/us)
		 */
		double _drift_velocity;

		/** The outer radius of the TPC (mm)
		 */
		double _tpc_outer_radius;

		/** The inner radius of the TPC (mm)
		 */
		double _tpc_inner_radius;

		/** The absolute value of the z coordinate of the outer position of the
		 *  TPC (mm)
		 */
		double _outer_z;

		/** The absolute value of the z coordinate of the inner position of the
		 *  TPC (mm)
		 */
		double _inner_z;


		/** temporary correction
		 */
		double _sigma;

	}; // /PhotodotReconstructionProcessor
  
} // /namespace marlintpc

#endif //PHOTODOTRECONSTRUCTIONPROCESSOR_H_

#endif //USE_LCCD
