#ifndef TRACK_FITTER_BASE_H
#define TRACK_FITTER_BASE_H

#include <string>
#include <marlin/StringParameters.h>
#include <EVENT/LCRunHeader.h>
#include <EVENT/Track.h>
#include <IMPL/TrackImpl.h>
#include <EVENT/TrackerHit.h>

namespace gear
{
    class TPCParameters;
    class PadRowLayout2D;
}//namespace gear

namespace marlintpc
{
	/** 
	 *   The TrackFitterBase is a virtual class from which the
	 *   actual track fitters are derived. It provides a common interface
	 *   which is accessed by the track fitter processors and analysis 
	 *   processors.                                                  <br>
	 *   Most members are purely virtual.
	 *   They have to be overloaded/implemented in the
	 *   derived classes.
	 *
	 *   To calculate something like a single point resolution it
	 *   is required that these "points" are accessed via the 
	 *   EVENT::TrackerHit as an interface. Some methods like the global
	 *   likelihood method don't use the xy-information from hits but
	 *   refit a x-position from the charge on the pads using the given 
	 *   track fit . This requires a calculateResiduals method
	 *   individually for each fitting method.
	 *   To have a common interface for all methods the TrackerHit
	 *   is used, through which the more low level information like 
	 *   pulses can be accessed.                                     <br>
	 *   The calculateResiduals method is not purely virtual. It is 
	 *   implemented for the generic case which calculates the distance
	 *   to the point of closest approach (pca) in the readout plane 
	 *   and the z-distance  to the pca in the s-z plane.
	 */

    class TrackFitterBase
    {
	public:

	/** Defines constants to determine which fitter type was used
	 *  to perform the fit and calculate the residuals.
	 *  When starting to implement a fitter please add a constant 
	 *  and commit this file to the trunk  so your constant is 
	 *  reserved.
	 */
	// These variables are static, don't forget to instantiate them in TrackFitterBase.cc
	struct FitterTypes
	{
	    //static const int BASE = 0;
	    static const unsigned char LIKELIHOOD = 1; ///< Global likelihood method
	    static const unsigned char SIMPLECHISQUARE = 2; ///< Simple \f$\chi^2\f$ without geometry information
	    static const unsigned char SIMPLECHISQUAREPADS = 3; ///< Simple \f$\chi^2\f$ using pad geometry information
	    static const unsigned char LINEARREGRESSION = 4; ///< Linear regression for a straight line
	    static const unsigned char KALMAN = 5; ///< Kalman filtering of the track
	    static const unsigned char SIMPLEMINIMIZER = 6; ///< Fitter which does not need parameters
	    static const unsigned char CHISQUARE = 7; ///< The \f$\chi^2\f$ fitter
	};

	static const unsigned char FITFAILEDBIT  =  30;  ///< Bit 30 in the TrackType word is fit failed
	//	static const int FITFAILED = 0x1 << FITFAILEDBIT;///< The mask to test if the fit failed

	/** The Constructor. The fit start values are initialised here and the flags whether to use the
	 *  are set to false (start values are taken from seed track).
	 */
	TrackFitterBase();

	virtual ~TrackFitterBase(){}

	/// Get the revision of the actual fitter implementation.
	/// This is needed for revision logging.
        /// Implementation should look like:
        /// \verbatim return std::string( "$Rev$" ); \endverbatim
	/// In case you don't reimplement the calculateResidulas methods you 
	/// should also state the revision of this base class:
        /// \verbatim return std::string( "$Rev$ ; TrackFitterBase ") + TrackFitterBase::getRevision(); \endverbatim
        /// Don't forget to switch on the 'svn Rev' property for this file:
	/// \verbatim svn propset svn:keywords Rev TrackFitterImpl.cc \endverbatim
	virtual std::string getRevision() const;

	/// Information what fitter implementation this is.
	virtual unsigned char getFitterType() const = 0;

	/** The function which does the actual fitting.
	 *  The returned TrackImpl is created from heap. Be careful to 
	 *  either add it to an event/collection or to delete it otherwise
	 *  to avoid memory leaks.
	 */
	virtual IMPL::TrackImpl *fitTrack(EVENT::Track const * seedTrack) const = 0;

	/** Return the distance of the hit \c testHitNumber of the testTrack 
	 *  to the point of closest approach on the referenceTrack wrt.\ the 
	 *  hit.
	 *  If no referenceTrack is given the
	 *  distance to the testTrack is returned.
	 * 
	 *  Throws EVENT::DataNotAvailableException in case there is no hit number 
	 *  "testHitNumber" on testHit.
	 *
	 *  Calls  calculateResidual(EVENT::TrackerHit *testHit, EVENT::Track *trackWithoutTestHit)
	 *  in this implementation.
	 */
	virtual EVENT::DoubleVec calculateResiduals(EVENT::Track const * testTrack,
					     unsigned int testHitNumber,
					     EVENT::Track const * referenceTrack = NULL) const;

	/** Return the distance of the given hit to the test track 
	 *  where this hit has been removed.
	 *  The residual is calculated as the distance 
	 *  to the point of closest approach on the referenceTrack wrt.\ the  hit.
	 */
	virtual EVENT::DoubleVec calculateResiduals(EVENT::TrackerHit const * testHit,
						    EVENT::Track const *trackWithoutTestHit) const;
	
	
	/** Convenience function to set bits 16-23 to be the track type.
	 */
	static void setTrackFitterType(unsigned char type, IMPL::TrackImpl *track);

	/** Convenicene function to get the tracker type out of the trackType word
	 */
	static unsigned char getFitterTypeFromTrack(int trackTypeWord);
	
	/** Set a d0 start value for the fit. This overwrites the value from the seed track.
	 */
	void setD0Start(double d0);

	/** Set a phi start value for the fit. This overwrites the value from the seed track.
	 */
	void setPhiStart(double phi);

	/** Set a omega start value for the fit. This overwrites the value from the seed track.
	 */
	void setOmegaStart(double omega);

	/** Set a tanLambda start value for the fit. This overwrites the value from the seed track.
	 */
	void setTanLambdaStart(double tanLambda);

	/** Set a z0 start value for the fit. This overwrites the value from the seed track.
	 */
	void setZ0Start(double z0);

	/** Switch to turn off the fitting in z. Fit is only performed in x-y / r-phi plane.
	 *  As default noZFit is false, the fit is performed in three dimensions.
	 */
	void setNoZFitFlag(bool noZFit = true);

	/** Get the noZFit flag
	 */
	bool getNoZFitFlag() const;

	/** Fix or release d0
	 */
	void fixD0(bool fix = true);

	/** Fix or release phi
	 */
	void fixPhi(bool fix = true);

	/** Fix or release omega
	 */
	void fixOmega(bool fix = true);

	/** Fix or release tanLambda
	 */
	void fixTanLambda(bool fix = true);

	/** Fix or release z0
	 */
	void fixZ0(bool fix = true);

    protected:

	/// Calculate the angular distance on a circle.
	/// The maximum can be \f$\pi\f$.
	double distanceOnCircle(double phi1, double phi2) const;

	/// The start value for the fit and a flag whether to apply d0Start 
	/// or the value from the seed track
	double _d0Start; bool _useD0Start;

	/// The start value for the fit and a flag whether to apply phiStart 
	/// or the value from the seed track
	double _phiStart; bool _usePhiStart;

	/// The start value for the fit and a flag whether to apply omegaStart 
	/// or the value from the seed track
	double _omegaStart; bool _useOmegaStart;

	/// The start value for the fit and a flag whether to apply tanLambdaStart 
	/// or the value from the seed track
	double _tanLambdaStart; bool _useTanLambdaStart;

	/// The start value for the fit and a flag whether to apply z0Start 
	/// or the value from the seed track
	double _z0Start; bool _useZ0Start;
	
        /// flag whether to perform a 2D or a 3D fit
	bool _noZFit;

	/// flags whether to fix a fit parameter
	bool _fixD0, _fixPhi, _fixOmega, _fixTanLambda, _fixZ0;
    };

}// namespace marlintpc

#endif // TRACK_FITTER_BASE_H
