#ifndef TRACKMERGERPROCESSOR_H
#define TRACKMERGERPROCESSOR_H 1

#include "marlin/Processor.h"
#include "lcio.h"
#include <string>
#include <map>

#include <EVENT/TrackerHit.h>

namespace marlintpc
{

/** This is a processor for TrackFitterKalmanProcessor for Multi Module.
 *
 *  @author Daizo Sakuragi, Kinki University
 *  @author Keisuke Fujii, KEK
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A collection of track candidates (lcio::Tracks) containing the TrackerHits which
 *      is merged Hits between modules.<br>
 *	Note: The LCIO::TRBIT_HITS flag has to be set when this collection is created so
 *      the hits are stored in the track.
 *
 *  <h4>Output</h4>
 *  \li Collection of merged Tracks
 *
 *  @param InputTrackCandidates The the name of the input collection of track candidates (default: TPCTrackCandidates)
 *  @param OutputMergedTracks The name of the collection of the merged tracks (default: TPCMergedTracks)
 *  @param SetOutputTransient If not 0 the output collection is set transient (default: 0)
 */
class TrackMergerProcessor : public marlin::Processor {
  
public:

	/** Default constructor
	*/
	TrackMergerProcessor();


	/** Destructor
	*/
	~TrackMergerProcessor();


	/** Return a new instance of this processor
	*/
       	Processor * newProcessor();


	/** Called at the begin of the job before anything is read.
	*/
	void init();


	/** Called once per run to process the event's header
	*/
	void processRunHeader (EVENT::LCRunHeader *run);


	/** Called for every event - the working horse.
	*/
	virtual void processEvent(EVENT::LCEvent *evt);


	/** Called for every event - right after processEvent() has been called for all processors.
	*/
	virtual void check (EVENT::LCEvent *evt);


	/** Called after data processing for clean up in the inverse order of the init() method so that resources allocated in the first processor also will be available for all following processors.
	*/
	virtual void end();

private:

	/** the name of the input collection of track candidates
	*/
	std::string _input_track_candidates_collection_name;

	/** the name of the collection the merge tracks
	*/
	std::string _output_merged_tracks_collection_name;

	/** the TrackerHits with ModuleID
	 */
       	std::multimap<int, IMPL::TrackImpl*> *_tracksByModuleID;

	/** Give the status which is set to the transient flag of the output collection.
	 *  It is an int instead of a boolean, because the processor parameter cannot be a boolean.
	 *  Hits and tracks can be set separately.
	 */
	int _outputIsTransient;    //< if not 0 the hits collection is set transient (default 0)

};

}// namespace marlintpc

#endif // TRACKMERGERPROCESSOR_H



