#include <iostream>
#include <stdexcept>
#include <vector>

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

#include "ChannelMapperProcessor.h"
#include "LCObjectCopier.h"

#include <marlin/Exceptions.h>

using namespace lcio ;
using namespace marlin ;
using namespace std ;

namespace marlintpc
{

ChannelMapperProcessor aChannelMapperProcessor;

ChannelMapperProcessor::ChannelMapperProcessor() : Processor("ChannelMapperProcessor")
{

    // the processor description
    _description = "ChannelMapperProcessor changes cellID0 to map from hardware channel numbers to sowftware/logical channel numbers." ;

    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(LCIO::TRACKERPULSE,
                            "InputCollectionName" ,
                            "Name of the input collection"  ,
                            _inputColName ,
                            std::string("TPCPulses")) ;

    registerProcessorParameter("MappingCollectionName",
                               "Name of the collection with the mapping",
                               _mappingColName,
                               std::string("TPCChannelMapping"));

    registerOptionalParameter("OutputCollectionName" ,
                              "Optional: Name of output collection"  ,
                              _outputColName ,
                              std::string("TPCPulsesMapped")) ;

    registerProcessorParameter("WriteOutputToStorage",
                               "if true the output collection is written",
                               _outputIsPersistent,
                               true);
}

void ChannelMapperProcessor::processRunHeader(LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2556 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
      {
	if (! i->second->isOptional() || i->second->valueSet())
	  {
	    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
	  }
      }
}


void ChannelMapperProcessor::processEvent(LCEvent * evt)
{
    // some status output
    if (evt->getEventNumber() % 1000 == 0)
        cout << "\r" << _processorName << " processing event " << evt->getEventNumber() << flush;

    // get input collection and check if there is data in it
    // We cast the input to LCCollectionVec if we want to use the iterator or modify it
    LCCollectionVec *inputColl = NULL;
    try
        {
            inputColl = dynamic_cast<LCCollectionVec *>(evt->getCollection(_inputColName));

            if (inputColl->size() == 0)
                throw DataNotAvailableException(std::string("Empty pulses collection ") + _inputColName);
        }
    catch (DataNotAvailableException &e)
        {
            cout << _processorName << " : Empty event " << evt->getEventNumber() << endl;
            //      throw marlin::SkipEventException( this);
            return;
        }

    // the pulses collection is the "output"
    // if the input collection is writeable and no output colection name
    // is given we use the input collection
    LCCollectionVec *pulsesColl = NULL;

    // test if the collection is writeable
    try
        {
            dynamic_cast<TrackerPulseImpl*>(inputColl->getElementAt(0))->setCellID0(
                dynamic_cast<TrackerPulse*>(inputColl->getElementAt(0))->getCellID0());
        }
    catch (ReadOnlyException &e)
        {
            //cout << "collection is read only!" << endl;// debug

            // check if output collection name is given
            if (! parameterSet("OutputCollectionName"))
                {
                    cerr << "Error: collection is not writeable." << endl;
                    cerr << "You have to specify an OutputCollectionName." << endl;
                    cerr << "Remember to adapt the input collection name of the next processor!" << endl;
                    throw lcio::ReadOnlyException(_inputColName + " not writeable");
                }
            // debug:
            //      else
            //    cout << "will use output collection name" << endl;
        }

    // clone input if OutputCollectionName is given
    // (note that this is mandatory if input is writeOnly)
    if (parameterSet("OutputCollectionName"))
        {
            // cout << "cloning the input collection" << endl; //debug
            pulsesColl = new LCCollectionVec("TrackerPulse");
            for (LCCollectionVec::iterator it = inputColl->begin(); it < inputColl->end() ; it++)
                {
                    pulsesColl->addElement(LCObjectCopier::copy(dynamic_cast<TrackerPulseImpl*>(*it)));
                }

            // set the transient flag as defined by the processor parameter
            pulsesColl->setTransient(!_outputIsPersistent);
        }
    else
        pulsesColl = inputColl;

    // set the flags that cellID1 should be stored
    // it's safe to do it here, collection is writeable now
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_CM) ;
    pulsesColl->setFlag(trkFlag.getFlag());


    LCCollection *mappingCollection = evt->getCollection(_mappingColName);

    myMappingHandler.setChannelMap(mappingCollection);

    // sort the pulses according to CellID
//     for (LCCollectionVec::iterator it = pulsesColl->begin();
//             it < pulsesColl->end(); it++)
//         {
//             TrackerPulseImpl *pulse =
//                 dynamic_cast<TrackerPulseImpl*>(*it);

    std::vector<int> badChannels ;

    int nPulses = pulsesColl->getNumberOfElements() ;

    for( int i=0 ; i<nPulses ; ++i ) {
      
      TrackerPulseImpl* pulse = dynamic_cast<TrackerPulseImpl*>( pulsesColl->getElementAt(i) ) ;

            // debug
            /*
            cout << "pulse is " << pulse << endl;
            cout << "cellID is "<<  pulse->getCellID0() <<endl;
            cout << "padID is " <<  myMappingHandler.getPadID( pulse->getCellID0() ) << endl;

            */

            // the softwareID consists of two ints.
            // the first is the pad index, the second the module id
           try{
             std::pair<int, int> softwareID = myMappingHandler.getSoftwareID(pulse->getCellID0(),
                                                                             pulse->getCellID1());
             pulse->setCellID0(softwareID.first);
             pulse->setCellID1(softwareID.second);

           } catch( lcio::DataNotAvailableException& e) {

//                streamlog_out( ERROR ) << "************************************************************************" << std::endl 
//                                       << "*   error in channel mapping: " << std::endl 
//                                       << "*    " <<   e.what() << std::endl 
//                                       << "*   will skip current event ! " << std::endl 
//                                       << "************************************************************************" 
//                                       << std::endl ;
// 	       throw  marlin::SkipEventException( this ) ;

	     

	      // remove the channel from the output and increase counter:

	     streamlog_out( DEBUG2 )  << "************************************************************************"  << std::endl
				      << "*   will remove channel [" << pulse->getCellID0() <<"," << pulse->getCellID1() <<"] "
				      << "*   from collection !  " << std::endl 
				      << "************************************************************************" << std::endl ;

	     badChannels.push_back( i ) ;

	     _cMap[ std::make_pair( pulse->getCellID0() , pulse->getCellID1() )  ] ++ ;  
	     
           } 
    }
    // remove elements now - need to start from end of vector !!
    for( std::vector<int>::reverse_iterator it=badChannels.rbegin() ; it != badChannels.rend() ; ++it ){
      
      pulsesColl->removeElementAt( *it ) ; 
	     
    }


    // if the pulses collection has been created by cloning
    // we have to add it to the event
    if (parameterSet("OutputCollectionName"))
        {
            evt->addCollection(pulsesColl, _outputColName);
        }
//   _nEvt ++ ;
}

void ChannelMapperProcessor::end()
{


  if( _cMap.size() ) {
    
    streamlog_out( WARNING3 ) << "******************************************************************************" << std::endl
			      << "*  The following hardware channels have been dropped from the output          " << std::endl   
			      << "*  collection, due to unknown channel mapping (possibly unconnected channels) " << std::endl 
			      << "*  cellID0 \t cellID1 \t count (#events)                                      " << std::endl 
			      << "*                                                                             " << std::endl;
    
    for( MAP::iterator it=_cMap.begin() ; it != _cMap.end() ; ++it){
      
      streamlog_out( WARNING3 ) << "*   "  <<  it->first.first << "\t" << it->first.second << "\t" << it->second  << std::endl ;
    }
    
    streamlog_out( WARNING3 ) << "*                                                                              " << std::endl
			      << "*******************************************************************************" << std::endl ;
  }
}

}// namespace marlintpc
