
//*****************************************************
// Part of program to simulate a TPC with GEM readout *
//                                                    *
// Author: Astrid Muennich                            *
//                                                    *
//*****************************************************

//Programm to create a primary electrons along the particle track. 

#include "PrimaryIonisationProcessor.h"

//C++
#include <iostream>
#include <cmath>
#include <cstdlib>

//LCIO
#include <IO/LCWriter.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <EVENT/LCParameters.h>
#include <IMPL/SimTrackerHitImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>


//MARLIN
#include <marlin/Global.h>

//GEAR
#include <gearimpl/RectangularPadRowLayout.h>
#include <gearimpl/FixedPadSizeDiskLayout.h>
#include <gear/TPCParameters.h>

//CLHEP
#include <CLHEP/Random/RandExponential.h>

using namespace lcio ;
using namespace marlin ;
using namespace std;

#define NANO_C 0.299792458 // speed of light / Giga
#define MEAN_IONISIATION_ENERGY 26E-9 //26 eV for Argon in GeV

namespace marlintpc{

PrimaryIonisationProcessor aPrimaryIonisationProcessor ;

PrimaryIonisationProcessor::PrimaryIonisationProcessor() : Processor("PrimaryIonisationProcessor") {
  
  _description = "PrimaryIonisationProcessor creates electrons along mcparticle track ..." ;
  
  
  // register steering parameters: name, description, class-variable, default value

  registerInputCollection(    LCIO::MCPARTICLE,
                              "InputCollectionName" , 
			      "Name of the input collection with MCParticles"  ,
			      _inputCollectionName ,
			      string("MCParticle") ) ;

  registerOutputCollection(   LCIO::SIMTRACKERHIT ,
                              "OutputCollectionName" ,
                              "Name of the output collection with primary electrons" ,
                              _outputCollectionName ,
                              string("PrimaryElectrons") ) ;

  registerProcessorParameter( "ClusterInfo" , 
			      "File with Information about #e- in Cluster"  ,
			       clusterfile,
			      std::string("clusterinfofile")) ;

   registerProcessorParameter( "FullCurler" , 
			      "switch on simulation of low momentum particles "  ,
			      Curler,
			      (int) 1) ;

  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection"  ,
			      _transient,
			      (int) 0) ;

  registerProcessorParameter( "GroupElectrons" , 
			      "Flag whether to group electrons !!READ DOCU BEFORE TURNING ON!!), default: false"  ,
			      _groupElectrons,
			      false) ;
}


void PrimaryIonisationProcessor::init() { 

  streamlog_out(MESSAGE3) << _description << endl;

  ReadClusterInfoFile();

  //read TPC parameters from gear
  _BField         =Global::GEAR->getTPCParameters().getDoubleVal("BField") ;
  _TPCInnerRadius =Global::GEAR->getTPCParameters().getDoubleVal("TPCInnerRadius") ;
  _TPCRadius      =Global::GEAR->getTPCParameters().getDoubleVal("TPCRadius") ;
  _distance       =Global::GEAR->getTPCParameters().getMaxDriftLength() ;

  printParameters() ;

  _nRun = 0 ;
  _nEvt = 0 ;
  
  //set and initialize other private members
  //  r = gsl_rng_alloc(gsl_rng_mt19937);
  //parameters for HEED Parametrisation
  // for number of clusters per mm
  a0 = 41.1221      ;
  a1 = -30.6686     ;
  a2 = 22.839       ;
  a3 = -7.92683     ;
  a4 = 1.57031      ;
  a5 = -0.182929    ;
  a6 = 0.0116357    ;
  a7 = -0.000310538 ;
  clusterdistance=0.;
  numberofelectrons=0;
  numberofclusters=0.;
  InTPC=0;
  // coordinates on particle track
  xa=0.,ya=0.,za=0.;
  xv=0.,yv=0.,zv=0.;
  xend=0.,yend=0., zend=0., tend=0.;
  dx=0., dy=0., dz=0.,dr=0.; 

  dp=1.; 
  px=0., py=0., pz=0.,pt=0.,p=0., p_const=0.;
  energy=0.;
  mass=0.;
  charge=0.;
  pdg=0;
  delta_x=0.,delta_y=0.,delta_z=0.;


}

void PrimaryIonisationProcessor::processRunHeader( LCRunHeader* run) {


  //Revision logging
  run->parameters().setValue( _processorName + "_revision", "$Rev$" );

  //Parameter logging
  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

    if ( ! i->second->isOptional() || i->second->valueSet() ) 
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
  }
 
  _nRun++ ;
} 

void PrimaryIonisationProcessor::processEvent( LCEvent * evt ) { 
  
  LCCollection* mcParticles = 0;
  int nMCs=1;
  bool fullCurler=false;
  if(Curler)
    fullCurler=true;
  try {
    mcParticles = evt->getCollection( _inputCollectionName  ) ;
  } catch(EventException &exception){
    nMCs=0;
  }
  if(nMCs!=0){
    nMCs = mcParticles->getNumberOfElements() ;
    streamlog_out(DEBUG3) << nMCs << " MCParticles in file! " << endl;
  }
  else
    streamlog_out(DEBUG3) << "No objects in input file!" << endl;
  
  LCCollectionVec* primelVec = new LCCollectionVec( LCIO::SIMTRACKERHIT )  ;
  if( _transient )
    primelVec->setTransient(true);
  
  if( _nEvt == 0 ){
    // set simulation parameter. These will be needed by following processors
    primelVec->parameters().setValue("TPCRadius" , _TPCRadius ) ;
    primelVec->parameters().setValue("TPCInnerRadius" , _TPCInnerRadius ) ;
    primelVec->parameters().setValue("TPCLength" ,_distance ) ;
    primelVec->parameters().setValue("BField" , _BField ) ;
    primelVec->parameters().setValue("ClusterInfoFile" , clusterfile ) ;
  }
    
  InTPC=0;

  for(int j=0;j<nMCs;j++){
    MCParticle* mcpart = dynamic_cast<MCParticle*> (mcParticles->getElementAt(j) );
    SetParticleParameters(mcpart);
    bool isCurler=false;
    if(!fullCurler){

      if(sqrt(px*px+py*py)<1 && 
	 atan(sqrt(px*px+py*py)/(fabs(pz)))*180/M_PI > 88)
	isCurler=true;
    }

    if ((mcpart->getGeneratorStatus()<=1) && ((mcpart->getCharge())!=0)){

      InitializeTrackPoint();
      int e_curlercounter=0;
      int nElectronsInCurler=0; // count how many electrons habe been skipped in a curler
      // they are added to the next hit

      CalculateNumberOfClusters();
      // as long as cluster in TPC continue track
      while((xend+xv)*(xend+xv)+(yend+yv)*(yend+yv)<=_TPCRadius*_TPCRadius && 
	    fabs(zend+zv)<fabs(_distance) && p>0.0){
		
	if(energy-mass>0.){   
	  
	  CalculateClusterDistance();
	  
	  double delta_cut=0;
	  if(_BField==0){

	    CalculateTrackPointwithoutBfield();
	    delta_cut=60; // equals range of 100 um
	  }
	  if(_BField!=0){

	    CalculateTrackPointwithBfield();
	    delta_cut=1200; //equals 100 um at B=1T
	  }
	  if(dr>10 && dp>1) streamlog_out(WARNING0)<<"Warning: high dE/dx "<<endl; 
	  CalculateNumberOfElectrons();
		   
	  if(numberofelectrons > delta_cut) { //delta electron that moves away from track

	    double delta_range=0;
	    double delta_energy=MEAN_IONISIATION_ENERGY * numberofelectrons // energy loss
	                       - 17E-9 ;                      // minus the energy for ionisation itself in GeV
	    //angle to track in most cases 90 degrees
	    double delta_angle_track=acos(sqrt(delta_energy/energy));// cos of angle
	    double delta_angle_plane=2*M_PI*drand48();
	    
	    if(_BField==0)
	      delta_range=8.23935e-05*pow(numberofelectrons,1.7293); //mm
	    if(_BField!=0)
	      delta_range=(delta_energy/(0.3*_BField))*1000; //in mm
	    
	    double step=delta_range/(double)numberofelectrons;
	    
	    double xe=xend+xv;
	    double ye=yend+yv;
	    double ze=zend+zv;
	    
	    //direction if track would be in 0,1,0
	    double deltax=sin(delta_angle_track)*cos(delta_angle_plane);
	    double deltaz=cos(delta_angle_track);
	    double deltay=sin(delta_angle_track)*sin(delta_angle_plane);
	    //transform to track coordinates   
	    RotateToTrack(deltax,deltay,deltaz);
	    
	    
	    for(int enumber=0;enumber<numberofelectrons;enumber++)
	    {
			    
	      //Multiple Scattering
	      if(!(enumber%30) && !(enumber==0)){

		double scatter_phi=2*M_PI*drand48();
		double scatter_theta=1./6*M_PI*drand48();
		// new direction
		MultipleScattering(scatter_phi,scatter_theta);
	      }
			    
	      if(sqrt((xe+delta_x*step)*(xe+delta_x*step)+(ye+delta_y*step)*(ye+delta_y*step))>=_TPCInnerRadius){

		if(isCurler){
		  e_curlercounter++;
		  if (_groupElectrons)
		  {
		    nElectronsInCurler++;
		  }

		  if(e_curlercounter%10==0){
		    SimTrackerHitImpl* primel = new SimTrackerHitImpl ;
		    double pos[3]={xe+delta_x*step,ye+delta_y*step,ze+delta_z*step};
		    primel->setPosition(pos);
		    primel->setTime(tend);
		    primel->setMCParticle(mcpart);
		    if (_groupElectrons)
		    {
		      // when grouping is done don't put 10 electrons but the number
		      // of electrons since the last deposition
		      primel->setdEdx(nElectronsInCurler);
		      InTPC+=nElectronsInCurler;  // counter for # e-
		      nElectronsInCurler=0;
		    }
		    else
		    {
		      primel->setdEdx(MEAN_IONISIATION_ENERGY*10); // deposit the energy of 10 electrons
		      InTPC+=10;  // counter for # e-
		    }
		    primelVec->addElement( primel ) ;
		  }
		}
		else{
		  SimTrackerHitImpl* primel = new SimTrackerHitImpl ;
		  double pos[3]={xe+delta_x*step,ye+delta_y*step,ze+delta_z*step};
		  primel->setPosition(pos);
		  primel->setTime(tend);
		  primel->setMCParticle(mcpart);
		  primel->setdEdx(MEAN_IONISIATION_ENERGY); // 26 eV, the ionisation energy of 1 electron
		  primelVec->addElement( primel ) ;
		  InTPC++;  // counter for # e-
		}
		
	      }
	      xe+=delta_x*step;
	      ye+=delta_y*step;
	      ze+=delta_z*step;
	    }
	  } //delta e-
	  else{
	    if(sqrt((xend+xv)*(xend+xv)+(yend+yv)*(yend+yv))>=_TPCInnerRadius)
	    {
	      if (!_groupElectrons)
	      {
		for(int enumber=0;enumber<numberofelectrons;enumber++)
		{
		  if(isCurler) // particle is a curler, only deposit every 10th electron
		  {
		    e_curlercounter++;
		    if(e_curlercounter%10==0){
		      
		      SimTrackerHitImpl* primel2 = new SimTrackerHitImpl ;
		      double pos[3]={xend+xv,yend+yv,zend+zv};
		      primel2->setPosition(pos);
		      primel2->setTime(tend);
		      primel2->setMCParticle(mcpart);
		      primel2->setdEdx(MEAN_IONISIATION_ENERGY*10); // deposit the energy of 10 electrons
		      primelVec->addElement( primel2 ) ;
		      InTPC+=10;  // counter for # e-
		    }
		  }
		  else// no curler
		  { 
		    SimTrackerHitImpl* primel2 = new SimTrackerHitImpl ;
		    double pos[3]={xend+xv,yend+yv,zend+zv};
		    primel2->setPosition(pos);
		    primel2->setTime(tend);
		    primel2->setMCParticle(mcpart);
		    primel2->setdEdx(MEAN_IONISIATION_ENERGY); // 26 eV, the ionisation energy of 1 electron
		    primelVec->addElement( primel2 ) ;
		    InTPC++;  // counter for # e-
		  }//else curler
		}// loop electrons in this hit

	      }// if (!_groupElectrons)
	      else // group electrons, but take care that heavy curlers are treated corretly
	      {
		  bool skipHit = false;
		  if(isCurler)  // particle is curler
		  {
		     if ( ( (e_curlercounter+numberofelectrons) /10 - 
			    e_curlercounter /10                       ) == 0 ) 
		     // number of electrons has not reached the next decade
		     {
		       skipHit =  true; // skip the hit because not the 10th electron
		       // remember how many electrons have been skipped,
		       //just put them into the next hit
		       nElectronsInCurler+=numberofelectrons; 
		     }

		     e_curlercounter+=numberofelectrons;
		  }
		  
		  if (!skipHit) // write the hit 
		  {
		    SimTrackerHitImpl* primel2 = new SimTrackerHitImpl ;
		    double pos[3]={xend+xv,yend+yv,zend+zv};
		    primel2->setPosition(pos);
		    primel2->setTime(tend);
		    primel2->setMCParticle(mcpart);
		    // deposit the energy of numberOfElectrons (plus those which have been skipped)
		    primel2->setdEdx(MEAN_IONISIATION_ENERGY * (numberofelectrons+nElectronsInCurler) );
		    InTPC+=numberofelectrons+nElectronsInCurler;  // counter for # e-
		    nElectronsInCurler=0;
		    primelVec->addElement( primel2 ) ;
		  }

	      }// group electrons
	      
	    }// if hit is in TPC

	  }// else delta e-

	  CalculateEnergyLoss();
	  SetNextStartPoint();
	  CalculateNumberOfClusters();
	}
	else{//Energy lower than mass -> Particle stopped in TPC
	  break;}
		
      } //loop over clusters
      
    }//if particle ok
    else
    {
      // dump the properties of the rejected particle
      streamlog_out(DEBUG2) << "rejected particle ID " << mcpart->getPDG() <<" at "<<  mcpart->getVertex()[0] 
                                                                           <<" : " <<  mcpart->getVertex()[1]
                                                                           <<" : " <<  mcpart->getVertex()[2]
			    << "\t" //std::endl
			    << "p = " << mcpart->getMomentum() << " ; q = " << mcpart->getCharge()
			    << "\t" //std::endl
	                    << "genStat " << mcpart->getGeneratorStatus() 
	                    << " ; simStat " << mcpart->getSimulatorStatus() 
			    << std::endl;	
	
    }

  }// particle loop//                                                    *

    
  if (primelVec->empty() && _nEvt!=0) 
    delete primelVec;
  else
    evt->addCollection( primelVec , _outputCollectionName ) ;
  
  streamlog_out(DEBUG4) << "Deposited " << InTPC << " Electrons in event "
			<< evt->getEventNumber() << std::endl;
  //if(_nEvt%100==0)
  streamlog_out(MESSAGE3)<<"Event "<<_nEvt<<" done!"<<endl;
  
  _nEvt ++ ;
   
}


void PrimaryIonisationProcessor::SetParticleParameters(MCParticle* mcparticle){

  const double *v;
  const double *parray;
  energy=mcparticle->getEnergy();
  mass=mcparticle->getMass();
  pdg=mcparticle->getPDG();
  charge=mcparticle->getCharge();
  v=(mcparticle->getVertex());
  xv=v[0]; // Vx() is in mm
  yv=v[1]; 
  zv=v[2];  
  parray=mcparticle->getMomentum();
  px=parray[0];
  py=parray[1];
  pz=parray[2];
  pt=sqrt(px*px+py*py);
  p=sqrt(pt*pt+pz*pz);
  p_const=p;
}

void PrimaryIonisationProcessor::InitializeTrackPoint(){

  xa=0.; ya=0.; za=0.;
  xend=0.;yend=0.;zend=0.;
}

void PrimaryIonisationProcessor::CalculateTrackPointwithBfield(){

  // rotation angle
  double factor = charge * _BField * NANO_C;
  double alpha=factor/p*clusterdistance/1000.; //[1/mm]

  double omega = charge * NANO_C * _BField / pt /1000.; //curvature in 1/mm
  double s_prime = clusterdistance*pt/p; // track length in the projection

 // angle of the vector from the helix centre to the stat point of this step
  double psi = atan2( py, px ) + ( omega < 0 ? -1 : 1 ) * M_PI_2;

  // do not use the circle centre to avoid calculating small values from a big one.
//  // center of circle
//  double xm_rel=1000*py*dp/factor; //[mm]
//  double ym_rel=-1000*px*dp/factor;

  // this calculation is on very obvious, but it avoids using the helix centre.
  // Escpecially for small curverture this improves the precision since 
  // the difference of two large numbers is avoided.
  // For more comments see the residual calculation in TrackFitterBase
  
  xend = xa + ( cos( psi - omega*s_prime ) - cos(psi) ) / std::fabs( omega );
  yend = ya + ( sin( psi - omega*s_prime ) - sin(psi) ) / std::fabs( omega );
  zend = za + pz*dp/p*clusterdistance;
    
  // also rotate px and py so centre in the next step is correct
  double px_new=cos(alpha)*px + sin(alpha)*py;
  py = cos(alpha) * py - sin(alpha) * px;
  px = px_new;
}

void PrimaryIonisationProcessor::CalculateTrackPointwithoutBfield(){

  //coordinates on straight line along p
  dr=clusterdistance;
  dx=px/p_const;
  dy=py/p_const;
  dz=pz/p_const;
  xend=xa+dr*dx;
  yend=ya+dr*dy;
  zend=za+dr*dz;
}

void PrimaryIonisationProcessor::CalculateNumberOfClusters(){

  //from parametrisation one knows function (polynom) and constants
  double x=sqrt((energy*energy)/(mass*mass)-1);
  double logx = std::log(x);
  double log10000 = std::log(10000.0);
  if(x<10000)
    numberofclusters = a0+a1*logx+a2*logx*logx+a3*pow(logx,3)
			+a4*pow(logx,4)+a5*pow(logx,5)
			+a6*pow(logx,6)+a7*pow(logx,7);
  else
    numberofclusters = a0+a1*log10000+a2*log10000*log10000+a3*pow(log10000,3)
			+a4*pow(log10000,4)+a5*pow(log10000,5)
			+a6*pow(log10000,6)+a7*pow(log10000,7);
  // number of cluster per cm constant after p/m>10.000
}

void PrimaryIonisationProcessor::CalculateClusterDistance(){

  //  clusterdistance =10.*(gsl_ran_exponential(r,1./numberofclusters)); 
  clusterdistance =10.*(CLHEP::RandExponential::shoot(1./numberofclusters)); 
  //*10 because distance needed in mm
}

void PrimaryIonisationProcessor::CalculateNumberOfElectrons(){

  //pick a random value between 0 and 1 and check file if random number is bigger than
  //integrated probability
  double r=0.998177337261218*drand48(); 
  //cryptic factor because of cut off on distribution at n=7632 comming from HEED
  int i=0;
  while(r>electrons[i][1] && i<7632)
    i++;
  numberofelectrons=(int)electrons[i][0];
}

void  PrimaryIonisationProcessor::ReadClusterInfoFile(){

  int lines=7632;
  FILE* fin = fopen(clusterfile.c_str(),"r");
  if(! fin) {                             
       printf("File with cluster information could not be opened.\n");
       exit(1);                           
   }
   /* read in events */
   int events = 0;                  /* evetn counter */               
   while(!feof(fin)) {              /* until end of file read one line*/    
                                          
      fscanf(fin,"%lf %lf " ,
             &electrons[events][0],
	     &electrons[events][1]);
      events++;                            
      
      if(events >lines) {             /* check for array overflow */
         printf("More events than array range (%i) allows.\n", lines);
         break;                            
      }
   }
   
   printf("%i Events read. \n", events);
   fclose(fin);            /* close file */                
}

void PrimaryIonisationProcessor::CalculateEnergyLoss(){
  
  // assume an energy loss of 26 eV for each produced electron in argon
  energy=energy-numberofelectrons*MEAN_IONISIATION_ENERGY; 
  //  dp=(p-numberofelectrons*0.000000026)/p;
  dp = sqrt(energy*energy-mass*mass)/p;
  if(energy-mass<=0)
  {
    streamlog_out(DEBUG3)<<"Particle stopped in TPC "<< endl;
  }
  else
  {
    p=sqrt(energy*energy-mass*mass);
  }
}


void PrimaryIonisationProcessor::RotateToTrack(double x, double y, double z){

  double phi;
  if(px>=0)
    phi=atan(py/px);
  else
    phi=atan(py/px)+M_PI;
  if(px==0 && py<0)
    phi=-phi;
  double theta=acos(pz);
  delta_x=cos(phi)*cos(theta)*x-sin(phi)*y+cos(phi)*sin(theta)*z;
  delta_y=sin(phi)*cos(theta)*x+cos(phi)*y+sin(phi)*sin(theta)*z;
  delta_z=-sin(theta)*x+cos(theta)*z;
}

void  PrimaryIonisationProcessor::RotateToDelta(double x, double y, double z){

  double phi;
  if(delta_x>0)
    phi=atan(delta_y/delta_x);
  else
    phi=atan(delta_y/delta_x)+M_PI;
  if(delta_x==0 && delta_y<0)
    phi=-phi;
  double theta=acos(delta_z);
  delta_x=cos(phi)*cos(theta)*x-sin(phi)*y+cos(phi)*sin(theta)*z;
  delta_y=sin(phi)*cos(theta)*x+cos(phi)*y+sin(phi)*sin(theta)*z;
  delta_z=-sin(theta)*x+cos(theta)*z;
}

void  PrimaryIonisationProcessor::MultipleScattering(double phi, double theta){

  double x_s=sin(theta)*cos(phi);
  double z_s=cos(theta);
  double y_s=sin(theta)*sin(phi);
  RotateToDelta(x_s,y_s,z_s);
}


void PrimaryIonisationProcessor::SetNextStartPoint(){

  xa=xend;
  ya=yend;
  za=zend;
}

}//end of namespace marlintpc
