
// stl
#include <cassert>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/LCGenericObject.h>

// marlintpc
#include "Cylindrical2DMagneticField.h"

namespace marlintpc {

Cylindrical2DMagneticField aCylindrical2DMagneticField;

/** constructor
 */
Cylindrical2DMagneticField::Cylindrical2DMagneticField():
	MagneticField("Cylindrical2DMagneticField") {
}


/** destructor
 */
Cylindrical2DMagneticField::~Cylindrical2DMagneticField() {}


/** returns a new field... similar to how the Marlin processor registration works
*/
Field *Cylindrical2DMagneticField::new_field( void ) { return new Cylindrical2DMagneticField; }


/** constructs the field from an LCCollection of data objects (could be anything... field bin values or just parameters)
*/
bool Cylindrical2DMagneticField::construct( EVENT::LCCollection *collection ) {

	_field_map.clear();

	// check if type of field map is correct
	if (collection->getParameters().getStringVal("FieldType") != "Cylindrical2DMagneticField") {
		std::cerr << collection << "is no Cylindrical2DMagneticField" << std::endl;
		return false;
	}

	// get field map boundaries from the collection
	_r_min = collection->getParameters().getFloatVal("r_min");
	_r_max = collection->getParameters().getFloatVal("r_max");
	_z_min = collection->getParameters().getFloatVal("z_min");
	_z_max = collection->getParameters().getFloatVal("z_max");
	_r_bins = collection->getParameters().getIntVal("r_bins");
	_z_bins = collection->getParameters().getIntVal("z_bins");

	// get the field error
	_field_error.setRho(collection->getParameters().getFloatVal("field_error_r"));
	_field_error.setZ(collection->getParameters().getFloatVal("field_error_z"));

	// calculate bin distance, need to correct for first and last bin being on the edges of the map
	_dr = (_r_max - _r_min) / (_r_bins -1);
	_dz = (_z_max - _z_min) / (_z_bins -1);

	// create empty field map
	for (int z_bin = 0; z_bin < _z_bins; z_bin++ ) {
		_field_map.push_back(std::vector<FieldMapEntry>());
		for (int r_bin = 0; r_bin < _r_bins; r_bin++ ) {
			_field_map[z_bin].push_back(FieldMapEntry());
		}
	}

	for( int map_counter = 0; map_counter < collection->getNumberOfElements(); map_counter++ ) {

		EVENT::LCGenericObject *field_map_object = dynamic_cast<EVENT::LCGenericObject*>( collection->getElementAt(map_counter) );
		assert(field_map_object);

		FieldMapEntry map_entry;

		map_entry._Br = field_map_object->getFloatVal(0);
		map_entry._Bz = field_map_object->getFloatVal(1);

		// getting position in 2D array from position in 1D array
		int r_bin = map_counter % _r_bins;
		int z_bin = map_counter / _r_bins;

		_field_map[z_bin][r_bin] = map_entry;

	}

	return true;
}


/** gets the field value at a point given in coil coordinates
*/
CLHEP::Hep3Vector Cylindrical2DMagneticField::get_field_value_coil_coordinates( const CLHEP::Hep3Vector &position ) {

	CLHEP::Hep3Vector field_value = CLHEP::Hep3Vector();

	// check if requested position is within field map volume, otherwise return 0
	if (position.rho() < _r_min || position.rho() > _r_max || position.z() < _z_min || position.r() > _z_max) {
		return field_value;
	}

	const double raw_index_r = (position.rho() - _r_min) / _dr;			// the exact r position in units of bins (fractional)
	const double raw_index_z = (position.z() - _z_min) / _dz;			// the exact z position in units of bins (fractional)
	const unsigned int floor_index_r = static_cast<int>(raw_index_r); 			// the lower neighbouring bin (integer)
	unsigned int ceil_index_r = static_cast<int>(raw_index_r + 1.0);		// the higher neighbouring bin (integer)
	const unsigned int floor_index_z = static_cast<int>(raw_index_z); 			// the lower neighbouring bin (integer)
	unsigned int ceil_index_z = static_cast<int>(raw_index_z + 1.0);		// the higher neighbouring bin (integer)

	const double floor_weight_r = ceil_index_r - raw_index_r;	// how close are we to the lower bin? (one minus distance)
	const double ceil_weight_r = raw_index_r - floor_index_r;	// how close are we to the higher bin? (one minus distance)
	const double floor_weight_z = ceil_index_z - raw_index_z;	// how close are we to the lower bin? (one minus distance)
	const double ceil_weight_z = raw_index_z - floor_index_z;	// how close are we to the higher bin? (one minus distance)

	// check if we hit the upper bound and avoid accessing array out of bounds
	if ( ceil_weight_r == 0.0 ) ceil_index_r = floor_index_r;
	if ( ceil_weight_z == 0.0 ) ceil_index_z = floor_index_z;

	// 2D interpolation for Br
	const double Br = floor_weight_z * (floor_weight_r * _field_map[floor_index_z][floor_index_r]._Br + ceil_weight_r * _field_map[floor_index_z][ceil_index_r]._Br)
					+ ceil_weight_z * (floor_weight_r * _field_map[ceil_index_z][floor_index_r]._Br + ceil_weight_r * _field_map[ceil_index_z][ceil_index_r]._Br);

	// 2D interpolation for Bz
	const double Bz = floor_weight_z * (floor_weight_r * _field_map[floor_index_z][floor_index_r]._Bz + ceil_weight_r * _field_map[floor_index_z][ceil_index_r]._Bz)
					+ ceil_weight_z * (floor_weight_r * _field_map[ceil_index_z][floor_index_r]._Bz + ceil_weight_r * _field_map[ceil_index_z][ceil_index_r]._Bz);

	field_value.setRhoPhiZ(Br, position.phi() ,Bz);

	return field_value;
}

} // namespace marlintpc
