#include <stdexcept>
#include <iostream>
#include <cmath>
#include <complex>
using namespace std;

#include <streamlog/streamlog.h>

#include "IntersectionCalculator.h"

namespace marlintpc
{

std::pair<double,double> IntersectionCalculator::circle_circle(double R, double centreX, double centreY, double r)
{
    // M2 is the square of the distance of the circle centre from the origin
    double M2=centreX*centreX+centreY*centreY;

    // check that circles really intesect, throw exception otherwise
    if ( ( std::sqrt(M2)  > ( R + r ) ) || // centres are further away than sum of radii, 
	                                   // circles are completely separated
	 ( std::sqrt(M2) < fabs ( r -  R ) ) ) // difference in radii is larger than distance between
                                              // centres => one circle is completely inside the other
	throw NoIntersectionException("Circles don't intersect.");

    
    double K = r*r - R*R - M2;
    //cout << "K " <<K<< endl;

    double tmp1 = centreX /2./ M2 *K ;
    double tmp2 = std::sqrt( tmp1*tmp1 - K*K/4./M2 + centreY*centreY*R*R/M2);

    double x1= -tmp1 + tmp2;
    double x2= -tmp1 - tmp2;
    //cout << "x1 " <<x1<< endl;
    //cout << "x2 " <<x2<< endl;
    
    
    double tmp3=centreY /2./ M2 *K ;
    double tmp4 = std::sqrt( tmp3*tmp3 - K*K/4./M2 + centreX*centreX*R*R/M2);

    double y1= -tmp3 + tmp4;
    double y2= -tmp3 - tmp4;
    //cout << "y1 " <<y1<< endl;
    //cout << "y2 " <<y2<< endl;

    double point1[2], point2[2];
    point1[0]=x1;
    point2[0]=x2;
    // find correct pairing
    // take into accout rounding errors: Double has a precision of O(1e-15) 
    // Unfortunately the accuracy of handling large raddi is not very good, only ~1e-12
    // So the difference incl. rounding should be 0 better than 1e-10 reltaive to the members of the sum,
    // in this case r^2 or R^2
    if  ( ( fabs(x1*x1 +y1*y1 - R*R) /R /R  < 1e-10)
	  &&  ( fabs((x1-centreX)*(x1-centreX) +(y1-centreY)*(y1-centreY) - r*r)  / r /r  < 1e-10) )
    {
	point1[1]=y1;
	point2[1]=y2;
    }
    else // also test the other pairing 
      if ( ( fabs(x1*x1 +y2*y2 - R*R)  /R /R  < 1e-10)
	     &&  ( fabs((x1-centreX)*(x1-centreX) +(y2-centreY)*(y2-centreY) - r*r) /r/r  < 1e-10) )
	{
	    point1[1]=y2;
	    point2[1]=y1;
	}
	else // there is something wrong , throw an exception
	{
	  streamlog_out(DEBUG) << "fabs(x1*x1 +y1*y1 - R*R) = " << fabs(x1*x1 +y1*y1 - R*R) << endl
	       << "fabs((x1-centreX)*(x1-centreX) +(y1-centreY)*(y1-centreY) - r*r) = "
	       <<  fabs((x1-centreX)*(x1-centreX) +(y1-centreY)*(y1-centreY) - r*r)  << endl
	       << "fabs(x1*x1 +y2*y2 - R*R) = " << fabs(x1*x1 +y2*y2 - R*R) << endl
	       << "fabs((x1-centreX)*(x1-centreX) +(y2-centreY)*(y2-centreY) - r*r) = " 
	       <<  fabs((x1-centreX)*(x1-centreX) +(y2-centreY)*(y2-centreY) - r*r) << endl; 
	  streamlog_out(DEBUG) << " r = " << r << " ; R = " << R << endl;

	    std::cerr << "IntersectionCalculator::calculateIntersectionAngles: calculated points don't match with circles"
		 << std::endl;
	    throw std::logic_error("No intersection found");
	}
    
    //cout << point1[0] << "\t"<<point1[1] << "\t";
    //cout << point2[0] << "\t"<<point2[1] << endl;
    std::complex<double> point1c(point1[0],point1[1]);
    std::complex<double> point2c(point2[0],point2[1]);

    if (std::arg(point1c) < std::arg(point2c) )
	return std::pair<double,double>(std::arg(point1c),std::arg(point2c));
    else
	return std::pair<double,double>(std::arg(point2c),std::arg(point1c));
}

std::pair<std::pair<double,double>,std::pair<double,double> > IntersectionCalculator::circle_straightLine_xy(double R, double inclination, double xAxisIntersept)
{
    double denominator = inclination*inclination+1;

    double argumentOfSqrt = inclination*inclination*xAxisIntersept*xAxisIntersept / denominator*denominator
	- ( xAxisIntersept*xAxisIntersept - R*R )/denominator ;

    if ( argumentOfSqrt < 0)
	throw NoIntersectionException("Circle and straight line don't intersect.");

    double y1= - inclination* xAxisIntersept / denominator - std::sqrt(argumentOfSqrt);
    double y2= - inclination* xAxisIntersept / denominator + std::sqrt(argumentOfSqrt);

    streamlog_out(DEBUG2) << "calculated intersections: x="
			  <<horizontalLine_straightLine(y1,inclination,xAxisIntersept)
			  << "/"<<y1<<" ; "
			  <<horizontalLine_straightLine(y2,inclination,xAxisIntersept)
			  << "/"<<y2<<std::endl;

    return std::pair<std::pair<double,double>,std::pair<double,double> >
	(std::pair<double,double>(horizontalLine_straightLine(y1,inclination,xAxisIntersept),y1),
	 std::pair<double,double>(horizontalLine_straightLine(y2,inclination,xAxisIntersept),y2));
}

std::pair<double,double>  IntersectionCalculator::circle_straightLine_phi(double R, double inclination, double xAxisIntersept)
{
    streamlog_out(DEBUG2) << "IntersectionCalculator::circle_straightLine_phi("
			  <<  R<<","<< inclination<<", "<<xAxisIntersept<<")"<<std::endl;

    std::pair<std::pair<double,double>,std::pair<double,double> > pointsInXY
	= circle_straightLine_xy( R, inclination, xAxisIntersept);

    std::complex<double> point1c( pointsInXY.first.first, pointsInXY.first.second);
    std::complex<double> point2c( pointsInXY.second.first, pointsInXY.second.second);

    if (std::arg(point1c) < std::arg(point2c) )
	return std::pair<double,double>(std::arg(point1c),std::arg(point2c));
    else
	return std::pair<double,double>(std::arg(point2c),std::arg(point1c));
}

/// inclination with respect to y axis
std::pair<double,double>  IntersectionCalculator::horizonalLine_circle(double  y,  double centreX, double centreY, double r)
{
    streamlog_out(DEBUG2) << "IntersectionCalculator::horizonalLine_circle("
			  << y<<", "<< centreX<< ", "<< centreY <<", "<<  r<< std::endl;

    double argumentOfSqrt = r*r - (y - centreY) * (y - centreY);
	
    if ( argumentOfSqrt < 0)
	throw NoIntersectionException("Horizontal line and circle don't intersect.");

    return std::pair<double,double> (centreX - std::sqrt(argumentOfSqrt), 
				     centreX + std::sqrt(argumentOfSqrt));
}

double IntersectionCalculator::horizontalLine_straightLine(double  y,  double inclination, double xAxisIntersept)
{
    return inclination*y+xAxisIntersept;
}

} // namepsace marlintpc
