#ifndef TimePixParameters_h
#define TimePixParameters_h 1

//C++
#include <iostream>
#include <string>
#include <utility>

//LCIO
#include "lcio.h"
#include "EVENT/LCGenericObject.h"
#include "IMPL/LCGenericObjectImpl.h"

namespace tpcconddata
{

  /** The TimePix parameters.
   *  This class should be able to store all the relevant information from the dsc file.
   *  In principle all functions also for medipix are there, although all convenience functions
   *  (calling the DAC by name, not by index) are only available for the Timepix.
   */

  class TimePixParameters : public EVENT::LCGenericObject
  {

  public:
            
    /** Constructor needed to interpret LCCollection read from file/database.
     */
    TimePixParameters(EVENT::LCObject* obj);
    
    /** Constructor to create a TimePixParameters object directly
     *  By default parameters for one chip are created
     */
    TimePixParameters(int nChips=1);

    /** Important for memory handling*/
    virtual ~TimePixParameters();

    // the class interface:
    
    /// set the chip iID aka module number
    void setChipID (int chipID);
    /// get the chip iID aka module number
    int getChipID() const;
    
    /// Set the DACs for one chip, a 16 words long integer array 
    void setDACs(int chipIndex, int * DACs);
    // OK, this does not work since the getIntVal gives a copy, not a reference :-(
    // Get the complete array of DACs
    //const double * getDACs(int chipIndex) const; 

    /// Get a specific DAC by index
    int getDAC(int chipIndex, int dacIndex) const;
    
    /// set the chip type (3=timepix)
    void setType(int type);
    /// get the chip type (3=timepix) 
    int getType() const;
    
    /// set the polarity
    void setPolarity(int polarity);
    /// get the polarity
    int getPolarity() const;
    
    // get the DACs for the Timepix by name (convenience functions for not to use the index)
    int getIKrumm(int chipIndex) const;
    int getDisc(int chipIndex) const;
    int getPreamp(int chipIndex) const;
    int getBuffA(int chipIndex) const;
    int getBuffB(int chipIndex) const;
    int getDelayN(int chipIndex) const;
    int getTHLFine(int chipIndex) const;
    int getTHLCoarse(int chipIndex) const;
    int getTHHFine(int chipIndex) const;
    int getTHHCoarse(int chipIndex) const;
    int getFBK(int chipIndex) const;
    int getGnd(int chipIndex) const;
    int getTHS(int chipIndex) const;
    int getBiasVLDS(int chipIndex) const;
    int getRefLVDS(int chipIndex) const;
    
    /// Set the clock (in MHz)
    void setClock(double clock);
    /// Get the clock (in MHz)
    double getClock() const;
    
    /// Set the shutter end time (wrt. tigger)
    void setShutterEndTime(double shutterEndTime);
    /// Get  the shutter end time (wrt. tigger)
    double getShutterEndTime() const;

    /// Get the number of chips
    int getNChips() const;
    
    static std::string getRevision();
    static std::string getDefaultColName();
    
    void print(  std::ostream& os = std::cout ) const;
    
    // -------- need to implement abstract methods from LCGenericObject
    int getNInt() const;   
    int getNFloat() const; 
    int getNDouble() const;
    
    int getIntVal(int index) const;
    float getFloatVal(int index) const;
    double getDoubleVal(int index) const;

    // Objects of type TimePixParameters do not always have the same size, but
    // the size does not change at runtime. So it is fixed size.
    bool isFixedSize() const { return true; };

    const std::string getTypeName() const
    {
      return std::string("TimePixParameters");
    }
    const std::string getDataDescription() const
    {
      return std::string("i:ChipID,Type,Polarity,NChips,DACs[16NChips],d:clock(MHz),shutterEndTime(ns)");
    }
    
  protected:
    bool createObj;
    IMPL::LCGenericObjectImpl* myObj;
    
  };                           // end of class
  
    std::ostream &operator<<(std::ostream &os, const TimePixParameters &p);

}                                //end namespace tpcconddata
#endif
