#include "ChargeOnPadsVisualisationProcessor.h"

//AIDA
#ifdef MARLIN_USE_AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram2D.h>
#endif

//MARLIN
#include <marlin/Global.h>
#include <marlin/Exceptions.h>
	
//LCIO
#include <EVENT/LCCollection.h>

//C++
#include <cmath>


//MARLINTPC
#include "TPCVoxel.h"

//GEAR
#include <gear/TPCParameters.h>


using namespace std;
using namespace lcio; 
using namespace marlin;

namespace marlintpc{
  
  ChargeOnPadsVisualisationProcessor aChargeOnPadsVisualisationProcessor ;
  
  
  ChargeOnPadsVisualisationProcessor::ChargeOnPadsVisualisationProcessor() : Processor("ChargeOnPadsVisualisationProcessor"){
    
    _description = "ChargeOnPadsVisualisationProcessor makes 2D histograms of charges on pad plane..." ;
    
    
    //Register steering parameters: name, description, class-variable, default value
  
  
    registerInputCollection(    LCIO::LCGENERICOBJECT,
                                "InputCollectionName" , 
  			        "Name of the input collection with charge on pads (default: DistributedElectrons)"  ,
  			        _inputCollectionName,
  		  	        string("DistributedElectrons") ) ;

    registerProcessorParameter( "ResetHistogram" ,
				"Resets histogram after each event if turned on (0) (default: 0)",
				_resetHisto ,
				(int) 0);
  
  }
  
  void ChargeOnPadsVisualisationProcessor::init() { 
  
    streamlog_out(MESSAGE2) << _description << endl;
  
    //Usually a good idea....
    printParameters() ;
  
    //For internal bookkeeping.
    _nEvt=0;
    _nRun=0;
  
    //Get the pad layout from gear.
    _padLayout=&(Global::GEAR->getTPCParameters().getPadLayout());
  
    //Pad layout type.
    _padLayoutType=_padLayout->getPadLayoutType();
  
    unsigned int xBins=0;
    unsigned int yBins=0;
    double leftBorder  = 0;
    double rightBorder = 0;
    double upperBorder = 0;
    double lowerBorder = 0;

    //Set borders and bins according to the type of the padplane
    switch( _padLayoutType ){
    case gear::PadRowLayout2D::POLAR: {

      leftBorder  = -_padLayout->getPlaneExtent()[1];;
      rightBorder = -leftBorder;
      upperBorder = -leftBorder;
      lowerBorder = leftBorder;

      xBins = static_cast<int>( (2 * rightBorder) / _padLayout->getRowHeight(0) );
      yBins = xBins;

      break;
    }
    case gear::PadRowLayout2D::CARTESIAN: {

      for( int i=0; i < _padLayout->getNRows(); i++ ){
	if( xBins < _padLayout->getPadsInRow(i).size() )
	  xBins = _padLayout->getPadsInRow(i).size() ;
      }

      yBins = _padLayout->getNRows();

      leftBorder  = _padLayout->getPlaneExtent()[0];
      rightBorder = _padLayout->getPlaneExtent()[1];
      lowerBorder = _padLayout->getPlaneExtent()[2];
      upperBorder = _padLayout->getPlaneExtent()[3];

      break;
    }
    default:
      throw gear::Exception("Unknown pad layout type");
    };
  
    //The AIDA histogram.
    _chargeOnPads=AIDAProcessor::histogramFactory(this)
       ->createHistogram2D("ChargeOnPads","ChargeDeposition",xBins,leftBorder,rightBorder,
                                                             yBins,lowerBorder,upperBorder);
  }
  
  
  
   
  void ChargeOnPadsVisualisationProcessor::processRunHeader( LCRunHeader* run) {
  
    run->parameters().setValue( _processorName + "_revision", "$Rev: 971 $" );
  
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
      if ( ! i->second->isOptional() || i->second->valueSet() ) {
        run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
      }
    }
    _nRun++;
  
  } 
  
   
  void ChargeOnPadsVisualisationProcessor::processEvent( LCEvent * evt ) {
  
    LCCollection* hitsOnPads = 0;
  
    int entries=0;
  
    //Get collection with charges on pads
    try{ 
      hitsOnPads = evt->getCollection( _inputCollectionName );
      entries=hitsOnPads->getNumberOfElements();

    }catch( EventException &exception){

      //The collection of TPCVoxels wasn't in the event, skip this event
      throw marlin::SkipEventException( this );
    }

   //Clear histogram
    if( !_resetHisto )
      _chargeOnPads->reset();
  
    //Loop over all entries in input collection
    for( int i=0; i < entries; i++ ){
  
      TPCVoxel* myVoxel= new TPCVoxel( dynamic_cast<LCGenericObject*>( hitsOnPads->getElementAt( i ) ) );
  
      double xHit = 0;
      double yHit = 0;
  
      //Set hit coordinates dependently of the pad layout
      switch( _padLayoutType ){
      case gear::PadRowLayout2D::POLAR:{

        double radius=_padLayout->getPadCenter(myVoxel->getPadIndex())[0];
        double phi=_padLayout->getPadCenter(myVoxel->getPadIndex())[1];
        xHit=radius*cos(phi);
        yHit=radius*sin(phi);
        break;
      }
      case gear::PadRowLayout2D::CARTESIAN:{

        xHit=_padLayout->getPadCenter(myVoxel->getPadIndex())[0];
        yHit=_padLayout->getPadCenter(myVoxel->getPadIndex())[1];
        break;
      }
      default:
        throw("Unknown pad layout type!");
      };

      //Fill charge in histogram
      _chargeOnPads->fill( xHit, yHit, myVoxel->getCharge() );
 
      delete myVoxel;
    }
    //if( _nEvt % 100 == 0)
    //streamlog_out(MESSAGE3) << "Event " << evt->getEventNumber() << " done!" << endl;
    
    _nEvt++ ;
  }

  
} //end of namesspace marlintpc
