#include "HistogramTrackerDataProcessor.h"
#include <iostream>
#include <string>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerData.h>
#include <Exceptions.h>

using namespace lcio;
using namespace marlin;
using namespace std;

// Use AIDA histograms
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h> // or IHistogram2D

HistogramTrackerDataProcessor aHistogramTrackerDataProcessor;


HistogramTrackerDataProcessor::HistogramTrackerDataProcessor() : Processor("HistogramTrackerDataProcessor")
{
  // the processor description
  _description = "HistogramTrackerDataProcessor is intended to ...";
  
  /* register steering parameters: name, description, class-variable, default value

    for example:
  */

        registerInputCollection(    LCIO::TRACKERDATA,
            "InputCollectionName" ,
            "Name of the input TrackerData collection, default: TPCData"  ,
            _inputColName ,
            string("TPCData")) ;

  /*
        registerOutputCollection(   LCIO::TRACKERDATA,
          "OutputCollectionName" ,
          string("Name of the output TrackerData collection")+
          string("only the data related to a pulse will be stored")+
          string(\, default: TPCDataInPulses),
          _outputCollectionName ,
          string("TPCDataInPulses")) ;

        // It's nive to have the default value in the comment
        registerProcessorParameter( "MinPulsesPerHit" ,
	       "Minimum number of pulses for a valid hit, default: 3",
	       _minPulsesPerHit,
	       int(3));

        // Optional parameters should only be used when set
        // in the steering file.
        // The default value is there only for technical reasons,
        // it is meaningless.
        //
        // If properly implemented the cut is not applied when the parameter is not set:
        //   if (parameterSet("ChageCut")) { //apply the cut }
        //
        registerOptionalParameter( "ChargeCut" ,
            "If the charge is below this value the data will be ignored",
            _chargeCut,
           float(17.8));
  */

  registerProcessorParameter( "NBinsSumQ" ,
			      "Number of bins in the SumQ histogram, default: 100",
			      _nBinsSumQ,
			      int(100));

  registerProcessorParameter( "NBinsMaxQ",
			      "Number of bins in the MaxQ histogram, default: 1024",
			      _nBinsMaxQ,
			      int(1024));

  registerProcessorParameter( "NBinsLength",
			      "Number of bins in the length histogram, default: 50",
			      _nBinsLength,
			      int(50));

  registerProcessorParameter( "MaxSumQ",
			      "Maximum in the SumQ histogram, default: 5000.",
			      _maxSumQ,
			      float(5000.));
}

HistogramTrackerDataProcessor::~HistogramTrackerDataProcessor(){}

void HistogramTrackerDataProcessor::init()
{ 
  streamlog_out(DEBUG) << " init() called  " << endl;
  
  // you might want to list the used parameters by invoking
  printParameters();

_maxQHistogram =  AIDAProcessor::histogramFactory(this)->
                    createHistogram1D( "maxQ",
                                       "maxQ",
                                       _nBinsMaxQ, 0 , _nBinsMaxQ);
_sumQHistogram =  AIDAProcessor::histogramFactory(this)->
                    createHistogram1D( "sumQ",
                                       "sumQ",
                                       _nBinsSumQ, 0 , _maxSumQ);
_pulseLengthHistogram =  AIDAProcessor::histogramFactory(this)->
                    createHistogram1D( "pulseLength",
                                       "pulse length",
                                       _nBinsLength, 0 , _nBinsLength);
_lengthVsSumQHistogram =  AIDAProcessor::histogramFactory(this)->
  createHistogram2D( "lengthVsSumQ",
		     "pulse length vs. sumQ",
		     _nBinsLength, 0 , _nBinsLength,
		     _nBinsSumQ, 0 , _maxSumQ);
_lengthVsMaxQHistogram =  AIDAProcessor::histogramFactory(this)->
  createHistogram2D( "lengthVsMaxQ",
		     "pulse length vs. maxQ",
		     _nBinsLength, 0 , _nBinsLength,
		     _nBinsMaxQ, 0 , _nBinsMaxQ);
}


void HistogramTrackerDataProcessor::processRunHeader(LCRunHeader* run)
{
  // do something for each run
} 


void HistogramTrackerDataProcessor::processEvent(LCEvent* evt)
{ 
  // do something for each event, usually access the input collection:

  // use an LCCollectionVec instead of LCCollection so we can use iterators later
  LCCollectionVec* inputColVec = 0;

  // do a try / catch in case there is an empty event
  try
  {
    inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
  }
  catch (DataNotAvailableException &e)
  {
    // no input data, just return and process the next event
    return;
  }

  // loop the data collection and do something 
  for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
       inputObjectIter < inputColVec->end() ;
       inputObjectIter++ )
  {
    // dynamic cast the input object (LCObject) to what it really is e.g.
    TrackerData * myTrackerData = dynamic_cast<TrackerData *>( *inputObjectIter );

    size_t pulseLength =  myTrackerData->getChargeValues().size();


  
    double sumQ=0;
    double maxQ=0;
    for (FloatVec::const_iterator adcValueIter = myTrackerData->getChargeValues().begin();
	 adcValueIter < myTrackerData->getChargeValues().end(); adcValueIter++)
    {
      if (*adcValueIter == 0)
      {
	streamlog_out(WARNING) << "Faulty TrackerData, has 0 entries" << std::endl;
	--pulseLength;
      }
      sumQ += *adcValueIter;
      if (*adcValueIter > maxQ) maxQ=*adcValueIter;
    }

    _pulseLengthHistogram->fill( pulseLength );

    _maxQHistogram->fill( maxQ );
    _sumQHistogram->fill( sumQ );
    _lengthVsSumQHistogram->fill( pulseLength , sumQ );
    _lengthVsMaxQHistogram->fill( pulseLength , maxQ );
    //_lengthVsSumQHistogram->fill( myTrackerData->getChargeValues().size() , sumQ );
    //_lengthVsMaxQHistogram->fill( myTrackerData->getChargeValues().size() , maxQ );
  }
}


void HistogramTrackerDataProcessor::check(LCEvent* evt)
{ 
  // check what is needed here 
}


void HistogramTrackerDataProcessor::end()
{ 
  // do something at the end of the processing, e.g. cleaning up
}
